@@ -1,6 +1,5 @@
-# -*- coding: utf-8 -*-
 #
-# Copyright (c), 2016-2019, SISSA (International School for Advanced Studies).
+# Copyright (c), 2016-2020, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -11,11 +10,12 @@
 """
 This module contains namespace definitions for W3C core standards and namespace related classes.
 """
-from __future__ import unicode_literals
 import os
 import re
+from collections.abc import MutableMapping, Mapping
 
-from .compat import MutableMapping, Mapping
+from .exceptions import XMLSchemaValueError, XMLSchemaTypeError
+from .qnames import get_namespace, local_name
 
 ###
 # Namespace URIs
@@ -64,20 +64,7 @@ LOCATION_HINTS = {
 
 
 ###
-# Helper functions and classes
-
-NAMESPACE_PATTERN = re.compile(r'{([^}]*)}')
-
-
-def get_namespace(name):
-    if not name or name[0] != '{':
-        return ''
-
-    try:
-        return NAMESPACE_PATTERN.match(name).group(1)
-    except (AttributeError, TypeError):
-        return ''
-
+# Base classes for managing namespaces
 
 class NamespaceResourcesMap(MutableMapping):
     """
@@ -126,13 +113,25 @@ class NamespaceMapper(MutableMapping):
     :param namespaces: initial data with namespace prefixes and URIs.
     :param register_namespace: a two-arguments function for registering namespaces \
     on ElementTree module.
+    :param strip_namespaces: if set to `True` uses name mapping methods that strip \
+    namespace information.
     """
-    def __init__(self, namespaces=None, register_namespace=None):
+    def __init__(self, namespaces=None, register_namespace=None, strip_namespaces=False):
         self._namespaces = {}
         self.register_namespace = register_namespace
+        self.strip_namespaces = strip_namespaces
         if namespaces is not None:
             self._namespaces.update(namespaces)
 
+    def __setattr__(self, name, value):
+        if name == 'strip_namespaces':
+            if value:
+                self.map_qname = self.unmap_qname = self._local_name
+            elif getattr(self, 'strip_namespaces', False):
+                self.map_qname = self._map_qname
+                self.unmap_qname = self._map_qname
+        super(NamespaceMapper, self).__setattr__(name, value)
+
     def __getitem__(self, prefix):
         return self._namespaces[prefix]
 
@@ -187,7 +186,7 @@ class NamespaceMapper(MutableMapping):
                 prefix += '0'
         self._namespaces[prefix] = uri
 
-    def map_qname(self, qname):
+    def _map_qname(self, qname):
         """
         Converts an extended QName to the prefixed format. Only registered
         namespaces are mapped.
@@ -198,24 +197,23 @@ class NamespaceMapper(MutableMapping):
         try:
             if qname[0] != '{' or not self._namespaces:
                 return qname
+            namespace, local_part = qname[1:].split('}')
         except IndexError:
             return qname
+        except ValueError:
+            raise XMLSchemaValueError("the argument 'qname' has a wrong format: %r" % qname)
+        except TypeError:
+            raise XMLSchemaTypeError("the argument 'qname' must be a string-like object")
 
-        qname_uri = get_namespace(qname)
-        for prefix, uri in self._namespaces.items():
-            if uri != qname_uri:
-                continue
-            if prefix:
-                self._namespaces[prefix] = uri
-                return qname.replace(u'{%s}' % uri, u'%s:' % prefix)
-            else:
-                if uri:
-                    self._namespaces[prefix] = uri
-                return qname.replace(u'{%s}' % uri, '')
+        for prefix, uri in sorted(self._namespaces.items(), reverse=True):
+            if uri == namespace:
+                return '%s:%s' % (prefix, local_part) if prefix else local_part
         else:
             return qname
 
-    def unmap_qname(self, qname, name_table=None):
+    map_qname = _map_qname
+
+    def _unmap_qname(self, qname, name_table=None):
         """
         Converts a QName in prefixed format or a local name to the extended QName format.
         Local names are converted only if a default namespace is included in the instance.
@@ -227,39 +225,54 @@ class NamespaceMapper(MutableMapping):
         :return: a QName in extended format or a local name.
         """
         try:
-            if qname[0] == '{' or not self:
+            if qname[0] == '{' or not self._namespaces:
                 return qname
+            prefix, name = qname.split(':')
         except IndexError:
             return qname
-
-        try:
-            prefix, name = qname.split(':', 1)
         except ValueError:
+            if ':' in qname:
+                raise XMLSchemaValueError("the argument 'qname' has a wrong format: %r" % qname)
             if not self._namespaces.get(''):
                 return qname
             elif name_table is None or qname not in name_table:
                 return '{%s}%s' % (self._namespaces.get(''), qname)
             else:
                 return qname
+        except (TypeError, AttributeError):
+            raise XMLSchemaTypeError("the argument 'qname' must be a string-like object")
         else:
             try:
                 uri = self._namespaces[prefix]
             except KeyError:
                 return qname
             else:
-                return u'{%s}%s' % (uri, name) if uri else name
+                return '{%s}%s' % (uri, name) if uri else name
 
-    def transfer(self, other):
+    unmap_qname = _unmap_qname
+
+    @staticmethod
+    def _local_name(qname, *_args, **_kwargs):
+        return local_name(qname)
+
+    def transfer(self, namespaces):
+        """
+        Transfers compatible prefix/namespace registrations from a dictionary.
+        Registrations added to namespace mapper instance are deleted from argument.
+
+        :param namespaces: a dictionary containing prefix/namespace registrations.
+        """
         transferred = []
-        for k, v in other.items():
-            if k in self:
-                if v != self[k]:
+        for k, v in namespaces.items():
+            if k in self._namespaces:
+                if v != self._namespaces[k]:
                     continue
             else:
                 self[k] = v
             transferred.append(k)
+
         for k in transferred:
-            del other[k]
+            del namespaces[k]
 
 
 class NamespaceView(Mapping):
@@ -292,9 +305,6 @@ class NamespaceView(Mapping):
     def __eq__(self, other):
         return self.as_dict() == dict(other.items())
 
-    def copy(self, **kwargs):
-        return self.__class__(self, **kwargs)
-
     def as_dict(self, fqn_keys=False):
         if fqn_keys:
             return {
