@@ -1,6 +1,5 @@
-# -*- coding: utf-8 -*-
 #
-# Copyright (c), 2016-2019, SISSA (International School for Advanced Studies).
+# Copyright (c), 2016-2020, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -8,7 +7,6 @@
 #
 # @author Davide Brunato <brunato@sissa.it>
 #
-from __future__ import unicode_literals
 from elementpath import XPath2Parser, XPathContext, ElementPathError
 from elementpath.datatypes import XSD_BUILTIN_TYPES
 
@@ -64,6 +62,7 @@ class XsdAssert(XsdComponent, ElementPathMixin):
         return self.token is not None and (self.base_type.parent is None or self.base_type.built)
 
     def parse_xpath_test(self):
+        # FIXME: parser's variables filled with XSD type with next elementpath minor release
         if not self.base_type.has_simple_content():
             variables = {'value': XSD_BUILTIN_TYPES['anyType'].value}
         else:
@@ -87,17 +86,22 @@ class XsdAssert(XsdComponent, ElementPathMixin):
         except ElementPathError as err:
             self.parse_error(err, elem=self.elem)
             self.token = self.parser.parse('true()')
+        finally:
+            self.parser.variables.clear()
 
     def __call__(self, elem, value=None, source=None, namespaces=None, **kwargs):
-        if value is not None:
-            self.parser.variables['value'] = self.base_type.text_decode(value)
         if not self.parser.is_schema_bound():
             self.parser.schema.bind_parser(self.parser)
 
-        if source is None:
-            context = XPathContext(root=elem)
+        if value is not None:
+            variables = {'value': self.base_type.text_decode(value)}
         else:
-            context = XPathContext(root=source.root, item=elem)
+            variables = {'value': ''}
+
+        if source is None:
+            context = XPathContext(root=elem, variables=variables)
+        else:
+            context = XPathContext(root=source.root, item=elem, variables=variables)
 
         default_namespace = self.parser.namespaces['']
         if namespaces and '' in namespaces:
@@ -115,8 +119,7 @@ class XsdAssert(XsdComponent, ElementPathMixin):
     # For implementing ElementPathMixin
     def __iter__(self):
         if not self.parent.has_simple_content():
-            for e in self.parent.content_type.iter_elements():
-                yield e
+            yield from self.parent.content_type.iter_elements()
 
     @property
     def attrib(self):
