@@ -1,6 +1,5 @@
-# -*- coding: utf-8 -*-
 #
-# Copyright (c), 2016-2019, SISSA (International School for Advanced Studies).
+# Copyright (c), 2016-2020, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -11,16 +10,16 @@
 """
 This module contains classes for XML Schema attributes and attribute groups.
 """
-from __future__ import unicode_literals
 from decimal import Decimal
+from collections.abc import MutableMapping
 from elementpath.datatypes import AbstractDateTime, Duration
 
-from ..compat import MutableMapping, ordered_dict_class
+from ..compat import ordered_dict_class
 from ..exceptions import XMLSchemaAttributeError, XMLSchemaTypeError, XMLSchemaValueError
-from ..qnames import XSD_ANNOTATION, XSD_ANY_SIMPLE_TYPE, XSD_SIMPLE_TYPE, \
-    XSD_ATTRIBUTE_GROUP, XSD_COMPLEX_TYPE, XSD_RESTRICTION, XSD_EXTENSION, \
-    XSD_SEQUENCE, XSD_ALL, XSD_CHOICE, XSD_ATTRIBUTE, XSD_ANY_ATTRIBUTE, \
-    XSD_ASSERT, get_namespace, get_qname
+from ..qnames import XSD_ANY_SIMPLE_TYPE, XSD_SIMPLE_TYPE, XSD_ATTRIBUTE_GROUP, \
+    XSD_COMPLEX_TYPE, XSD_RESTRICTION, XSD_EXTENSION, XSD_SEQUENCE, XSD_ALL, \
+    XSD_CHOICE, XSD_ATTRIBUTE, XSD_ANY_ATTRIBUTE, XSD_ASSERT, get_namespace, \
+    get_qname, is_not_xsd_annotation
 from ..helpers import get_xsd_form_attribute
 from ..namespaces import XSI_NAMESPACE
 
@@ -61,12 +60,6 @@ class XsdAttribute(XsdComponent, ValidationMixin):
         if not hasattr(self, 'type'):
             raise XMLSchemaAttributeError("undefined 'type' for %r." % self)
 
-    def __repr__(self):
-        if self.ref is None:
-            return '%s(name=%r)' % (self.__class__.__name__, self.prefixed_name)
-        else:
-            return '%s(ref=%r)' % (self.__class__.__name__, self.prefixed_name)
-
     def __setattr__(self, name, value):
         if name == "type":
             if not isinstance(value, XsdSimpleType):
@@ -116,7 +109,8 @@ class XsdAttribute(XsdComponent, ValidationMixin):
 
             for attribute in ('form', 'type'):
                 if attribute in self.elem.attrib:
-                    self.parse_error("attribute %r is not allowed when attribute reference is used." % attribute)
+                    self.parse_error("attribute %r is not allowed when "
+                                     "attribute reference is used." % attribute)
 
             child = self._parse_child_component(self.elem)
             if child is not None and child.tag == XSD_SIMPLE_TYPE:
@@ -166,7 +160,8 @@ class XsdAttribute(XsdComponent, ValidationMixin):
                 if child is not None and child.tag == XSD_SIMPLE_TYPE:
                     self.parse_error("ambiguous type definition for XSD attribute")
                 elif child is not None:
-                    self.parse_error("not allowed element in XSD attribute declaration: %r" % child[0])
+                    self.parse_error("not allowed element in XSD attribute "
+                                     "declaration: %r" % child[0])
         elif child is not None:
             # No 'type' attribute in declaration, parse for child local simpleType
             xsd_type = self.schema.BUILDERS.simple_type_factory(child, self.schema, self)
@@ -184,7 +179,8 @@ class XsdAttribute(XsdComponent, ValidationMixin):
             if 'fixed' in attrib:
                 self.parse_error("'default' and 'fixed' attributes are mutually exclusive")
             if self.use != 'optional':
-                self.parse_error("the attribute 'use' must be 'optional' if the attribute 'default' is present")
+                self.parse_error("the attribute 'use' must be 'optional' "
+                                 "if the attribute 'default' is present")
             if not self.type.is_valid(attrib['default']):
                 msg = "'default' value {!r} is not compatible with the type {!r}"
                 self.parse_error(msg.format(attrib['default'], self.type))
@@ -216,14 +212,11 @@ class XsdAttribute(XsdComponent, ValidationMixin):
         if xsd_classes is None or isinstance(self, xsd_classes):
             yield self
         if self.ref is None and self.type.parent is not None:
-            for obj in self.type.iter_components(xsd_classes):
-                yield obj
+            yield from self.type.iter_components(xsd_classes)
 
     def data_value(self, text):
         """Returns the decoded data value of the provided text as XPath fn:data()."""
-        for result in self.iter_decode(text, validation='skip'):
-            return result
-        return text
+        return self.decode(text, validation='skip')
 
     def iter_decode(self, text, validation='lax', **kwargs):
         if not text and self.default is not None:
@@ -232,7 +225,7 @@ class XsdAttribute(XsdComponent, ValidationMixin):
         if self.fixed is not None:
             if text is None:
                 text = self.fixed
-            elif text == self.fixed or validation == 'skip':
+            elif text == self.fixed:
                 pass
             elif self.type.text_decode(text) != self.type.text_decode(self.fixed):
                 msg = "attribute {!r} has a fixed value {!r}".format(self.name, self.fixed)
@@ -241,20 +234,22 @@ class XsdAttribute(XsdComponent, ValidationMixin):
         for result in self.type.iter_decode(text, validation, **kwargs):
             if isinstance(result, XMLSchemaValidationError):
                 yield result
+                continue
             elif isinstance(result, Decimal):
                 try:
                     yield kwargs['decimal_type'](result)
                 except (KeyError, TypeError):
                     yield result
-                break
             elif isinstance(result, (AbstractDateTime, Duration)):
                 try:
                     yield result if kwargs['datetime_types'] is True else text
                 except KeyError:
                     yield text
+            elif isinstance(result, str) and result.startswith('{') and self.type.is_qname():
+                yield text
             else:
                 yield result
-                break
+            break
 
     def iter_encode(self, obj, validation='lax', **kwargs):
         for result in self.type.iter_encode(obj, validation):
@@ -379,9 +374,11 @@ class XsdAttributeGroup(MutableMapping, XsdComponent, ValidationMixin):
             assert isinstance(value, dict), 'A dictionary object is required.'
             for k, v in value.items():
                 if k is None:
-                    assert isinstance(value, XsdAnyAttribute), 'An XsdAnyAttribute instance is required.'
+                    assert isinstance(value, XsdAnyAttribute), \
+                        'An XsdAnyAttribute instance is required.'
                 else:
-                    assert isinstance(value, XsdAttribute), 'An XsdAttribute instance is required.'
+                    assert isinstance(value, XsdAttribute), \
+                        'An XsdAttribute instance is required.'
 
     def _parse(self):
         super(XsdAttributeGroup, self)._parse()
@@ -402,7 +399,7 @@ class XsdAttributeGroup(MutableMapping, XsdComponent, ValidationMixin):
                     self.schema.default_attributes = self
 
         attributes = ordered_dict_class()
-        for child in filter(lambda x: x.tag != XSD_ANNOTATION, elem):
+        for child in filter(is_not_xsd_annotation, elem):
             if any_attribute is not None:
                 if child.tag == XSD_ANY_ATTRIBUTE:
                     self.parse_error("more anyAttribute declarations in the same attribute group")
@@ -420,15 +417,17 @@ class XsdAttributeGroup(MutableMapping, XsdComponent, ValidationMixin):
             elif child.tag == XSD_ATTRIBUTE:
                 attribute = self.schema.BUILDERS.attribute_class(child, self.schema, self)
                 if attribute.name in attributes:
-                    self.parse_error("multiple declaration for attribute {!r}".format(attribute.name))
-                else:
+                    self.parse_error("multiple declaration for attribute "
+                                     "{!r}".format(attribute.name))
+                elif attribute.use != 'prohibited' or elem.tag != XSD_ATTRIBUTE_GROUP:
                     attributes[attribute.name] = attribute
 
             elif child.tag == XSD_ATTRIBUTE_GROUP:
                 try:
                     ref = child.attrib['ref']
                 except KeyError:
-                    self.parse_error("the attribute 'ref' is required in a local attributeGroup", elem)
+                    self.parse_error("the attribute 'ref' is required "
+                                     "in a local attributeGroup", elem)
                     continue
 
                 try:
@@ -441,7 +440,8 @@ class XsdAttributeGroup(MutableMapping, XsdComponent, ValidationMixin):
                     elif self.redefine is not None:
                         if attribute_group_qname == self.name:
                             if attribute_group_refs:
-                                self.parse_error("in a redefinition the reference to itself must be the first")
+                                self.parse_error("in a redefinition the reference "
+                                                 "to itself must be the first")
                             attribute_group_refs.append(attribute_group_qname)
                             attributes.update(self._attribute_group.items())
                             continue
@@ -449,7 +449,8 @@ class XsdAttributeGroup(MutableMapping, XsdComponent, ValidationMixin):
                             # May be an attributeGroup restriction with a ref to another group
                             if not any(e.tag == XSD_ATTRIBUTE_GROUP and ref == e.get('ref')
                                        for e in self.redefine.elem):
-                                self.parse_error("attributeGroup ref=%r is not in the redefined group" % ref)
+                                self.parse_error("attributeGroup ref=%r is not "
+                                                 "in the redefined group" % ref)
                     elif attribute_group_qname == self.name and self.xsd_version == '1.0':
                         self.parse_error("Circular attribute groups not allowed in XSD 1.0")
                     attribute_group_refs.append(attribute_group_qname)
@@ -464,14 +465,17 @@ class XsdAttributeGroup(MutableMapping, XsdComponent, ValidationMixin):
                                 if name not in attributes:
                                     attributes[name] = attr
                                 elif name is not None:
-                                    self.parse_error("multiple declaration for attribute {!r}".format(name))
+                                    self.parse_error("multiple declaration for attribute "
+                                                     "{!r}".format(name))
                                 else:
                                     attributes[None] = attributes[None].copy()
                                     attributes[None].intersection(attr)
 
                         elif self.xsd_version == '1.0':
-                            self.parse_error("Circular reference found between attribute groups "
-                                             "{!r} and {!r}".format(self.name, attribute_group_qname))
+                            self.parse_error(
+                                "Circular reference found between attribute groups "
+                                "{!r} and {!r}".format(self.name, attribute_group_qname)
+                            )
 
             elif self.name is not None:
                 self.parse_error("(attribute | attributeGroup) expected, found %r." % child)
@@ -496,17 +500,21 @@ class XsdAttributeGroup(MutableMapping, XsdComponent, ValidationMixin):
                         except ValueError as err:
                             self.parse_error(err)
                     elif not attr.is_restriction(base_attr):
-                        self.parse_error("Attribute wildcard is not a restriction of the base wildcard")
+                        self.parse_error("Attribute wildcard is not a restriction "
+                                         "of the base wildcard")
                     continue
                 if self.derivation == 'restriction' and attr.type.name != XSD_ANY_SIMPLE_TYPE and \
                         not attr.type.is_derived(base_attr.type, 'restriction'):
-                    self.parse_error("Attribute type is not a restriction of the base attribute type")
+                    self.parse_error("Attribute type is not a restriction "
+                                     "of the base attribute type")
                 if base_attr.use != 'optional' and attr.use == 'optional' or \
                         base_attr.use == 'required' and attr.use != 'required':
                     self.parse_error("Attribute %r: unmatched attribute use in restriction" % name)
                 if base_attr.fixed is not None and \
-                        attr.type.normalize(attr.fixed) != base_attr.type.normalize(base_attr.fixed):
-                    self.parse_error("Attribute %r: derived attribute has a different fixed value" % name)
+                        attr.type.normalize(attr.fixed) != \
+                        base_attr.type.normalize(base_attr.fixed):
+                    self.parse_error("Attribute %r: derived attribute "
+                                     "has a different fixed value" % name)
 
             if self.redefine is not None:
                 pass  # In case of redefinition do not copy base attributes
@@ -519,7 +527,8 @@ class XsdAttributeGroup(MutableMapping, XsdComponent, ValidationMixin):
                     continue
                 elif name not in attributes:
                     if attr.use == 'required':
-                        self.parse_error("Missing required attribute %r in redefinition restriction" % name)
+                        self.parse_error("Missing required attribute %r in "
+                                         "redefinition restriction" % name)
                     continue
                 if attr.use != 'optional' and attributes[name].use != attr.use:
                     self.parse_error("Attribute %r: unmatched attribute use in redefinition" % name)
@@ -532,7 +541,8 @@ class XsdAttributeGroup(MutableMapping, XsdComponent, ValidationMixin):
                 try:
                     next_pos = keys.index(name)
                 except ValueError:
-                    self.parse_error("Redefinition restriction contains additional attribute %r" % name)
+                    self.parse_error("Redefinition restriction contains "
+                                     "additional attribute %r" % name)
                 else:
                     if next_pos < pos:
                         self.parse_error("Wrong attribute order in redefinition restriction")
@@ -541,7 +551,8 @@ class XsdAttributeGroup(MutableMapping, XsdComponent, ValidationMixin):
             self.clear()
 
         self._attribute_group.update(attributes)
-        if None in self._attribute_group and None not in attributes and self.derivation == 'restriction':
+        if None in self._attribute_group and None not in attributes \
+                and self.derivation == 'restriction':
             wildcard = self._attribute_group[None].copy()
             wildcard.namespace = wildcard.not_namespace = wildcard.not_qname = ()
             self._attribute_group[None] = wildcard
@@ -551,7 +562,8 @@ class XsdAttributeGroup(MutableMapping, XsdComponent, ValidationMixin):
             for attr in self._attribute_group.values():
                 if attr.name is not None and attr.type.is_key():
                     if has_key:
-                        self.parse_error("multiple key attributes in a group not allowed in XSD 1.0")
+                        self.parse_error("multiple key attributes in a group "
+                                         "not allowed in XSD 1.0")
                     has_key = True
 
         elif self.parent is None and self.schema.default_attributes == self.name:
@@ -586,17 +598,15 @@ class XsdAttributeGroup(MutableMapping, XsdComponent, ValidationMixin):
         if self.ref is None:
             for attr in self.values():
                 if attr.parent is not None:
-                    for obj in attr.iter_components(xsd_classes):
-                        yield obj
+                    yield from attr.iter_components(xsd_classes)
 
     def iter_decode(self, attrs, validation='lax', **kwargs):
         if not attrs and not self:
             return
 
-        if validation != 'skip':
-            for k in filter(lambda x: x not in attrs, self.iter_required()):
-                reason = "missing required attribute: %r" % k
-                yield self.validation_error(validation, reason, attrs, **kwargs)
+        for k in filter(lambda x: x not in attrs, self.iter_required()):
+            reason = "missing required attribute: %r" % k
+            yield self.validation_error(validation, reason, attrs, **kwargs)
 
         kwargs['level'] = kwargs.get('level', 0) + 1
         use_defaults = kwargs.get('use_defaults', True)
@@ -618,18 +628,16 @@ class XsdAttributeGroup(MutableMapping, XsdComponent, ValidationMixin):
                     try:
                         xsd_attribute = self.maps.lookup_attribute(name)
                     except LookupError:
-                        if validation != 'skip':
-                            reason = "%r is not an attribute of the XSI namespace." % name
-                            yield self.validation_error(validation, reason, attrs, **kwargs)
+                        reason = "%r is not an attribute of the XSI namespace." % name
+                        yield self.validation_error(validation, reason, attrs, **kwargs)
                         continue
                 else:
                     try:
                         xsd_attribute = self[None]  # None key ==> anyAttribute
                         value = (name, value)
                     except KeyError:
-                        if validation != 'skip':
-                            reason = "%r attribute not allowed for element." % name
-                            yield self.validation_error(validation, reason, attrs, **kwargs)
+                        reason = "%r attribute not allowed for element." % name
+                        yield self.validation_error(validation, reason, attrs, **kwargs)
                         continue
             else:
                 if xsd_attribute.use == 'prohibited' and \
@@ -665,10 +673,9 @@ class XsdAttributeGroup(MutableMapping, XsdComponent, ValidationMixin):
         if not attrs and not self:
             return
 
-        if validation != 'skip':
-            for k in filter(lambda x: x not in attrs, self.iter_required()):
-                reason = "missing required attribute: %r" % k
-                yield self.validation_error(validation, reason, attrs, **kwargs)
+        for k in filter(lambda x: x not in attrs, self.iter_required()):
+            reason = "missing required attribute: %r" % k
+            yield self.validation_error(validation, reason, attrs, **kwargs)
 
         use_defaults = kwargs.get('use_defaults', True)
         additional_attrs = [(k, v) for k, v in self.iter_predefined(use_defaults) if k not in attrs]
@@ -686,18 +693,16 @@ class XsdAttributeGroup(MutableMapping, XsdComponent, ValidationMixin):
                     try:
                         xsd_attribute = self.maps.lookup_attribute(name)
                     except LookupError:
-                        if validation != 'skip':
-                            reason = "%r is not an attribute of the XSI namespace." % name
-                            yield self.validation_error(validation, reason, attrs, **kwargs)
+                        reason = "%r is not an attribute of the XSI namespace." % name
+                        yield self.validation_error(validation, reason, attrs, **kwargs)
                         continue
                 else:
                     try:
                         xsd_attribute = self[None]  # None key ==> anyAttribute
                         value = (name, value)
                     except KeyError:
-                        if validation != 'skip':
-                            reason = "%r attribute not allowed for element." % name
-                            yield self.validation_error(validation, reason, attrs, **kwargs)
+                        reason = "%r attribute not allowed for element." % name
+                        yield self.validation_error(validation, reason, attrs, **kwargs)
                         continue
 
             for result in xsd_attribute.iter_encode(value, validation, **kwargs):
