@@ -1,6 +1,5 @@
-# -*- coding: utf-8 -*-
 #
-# Copyright (c), 2016-2019, SISSA (International School for Advanced Studies).
+# Copyright (c), 2016-2020, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -8,18 +7,16 @@
 #
 # @author Davide Brunato <brunato@sissa.it>
 #
-from __future__ import unicode_literals
-
 from ..exceptions import XMLSchemaValueError
-from ..qnames import XSD_ANNOTATION, XSD_GROUP, XSD_ATTRIBUTE_GROUP, XSD_SEQUENCE, \
+from ..qnames import XSD_GROUP, XSD_ATTRIBUTE_GROUP, XSD_SEQUENCE, \
     XSD_ALL, XSD_CHOICE, XSD_ANY_ATTRIBUTE, XSD_ATTRIBUTE, XSD_COMPLEX_CONTENT, \
-    XSD_RESTRICTION, XSD_COMPLEX_TYPE, XSD_EXTENSION, XSD_ANY_TYPE, XSD_OVERRIDE, \
+    XSD_RESTRICTION, XSD_COMPLEX_TYPE, XSD_EXTENSION, XSD_ANY_TYPE, \
     XSD_SIMPLE_CONTENT, XSD_ANY_SIMPLE_TYPE, XSD_OPEN_CONTENT, XSD_ASSERT, \
-    get_qname, local_name
+    get_qname, local_name, is_not_xsd_annotation, is_xsd_override
 from ..helpers import get_xsd_derivation_attribute
 
 from .exceptions import XMLSchemaDecodeError
-from .xsdbase import XsdComponent, XsdType, ValidationMixin
+from .xsdbase import XSD_TYPE_DERIVATIONS, XsdComponent, XsdType, ValidationMixin
 from .assertions import XsdAssert
 from .attributes import XsdAttributeGroup
 from .simple_types import XsdSimpleType
@@ -109,13 +106,13 @@ class XsdComplexType(XsdType, ValidationMixin):
 
         if 'block' in elem.attrib:
             try:
-                self._block = get_xsd_derivation_attribute(elem, 'block', ('extension', 'restriction'))
+                self._block = get_xsd_derivation_attribute(elem, 'block', XSD_TYPE_DERIVATIONS)
             except ValueError as err:
                 self.parse_error(err, elem)
 
         if 'final' in elem.attrib:
             try:
-                self._final = get_xsd_derivation_attribute(elem, 'final', ('extension', 'restriction'))
+                self._final = get_xsd_derivation_attribute(elem, 'final', XSD_TYPE_DERIVATIONS)
             except ValueError as err:
                 self.parse_error(err, elem)
 
@@ -163,7 +160,9 @@ class XsdComplexType(XsdType, ValidationMixin):
 
             if content_elem is not elem[-1]:
                 k = 2 if content_elem is not elem[0] else 1
-                self.parse_error("unexpected tag %r after simpleContent declaration:" % elem[k].tag, elem)
+                self.parse_error(
+                    "unexpected tag %r after simpleContent declaration:" % elem[k].tag, elem
+                )
 
         elif content_elem.tag == XSD_COMPLEX_CONTENT:
             #
@@ -173,9 +172,8 @@ class XsdComplexType(XsdType, ValidationMixin):
                 if mixed is not self.mixed:
                     self.mixed = mixed
                     if 'mixed' in elem.attrib and self.xsd_version == '1.1':
-                        self.parse_error(
-                            "value of 'mixed' attribute in complexType and complexContent must be same"
-                        )
+                        self.parse_error("value of 'mixed' attribute in complexType "
+                                         "and complexContent must be same")
 
             derivation_elem = self._parse_derivation_elem(content_elem)
             if derivation_elem is None:
@@ -195,7 +193,9 @@ class XsdComplexType(XsdType, ValidationMixin):
 
             if content_elem is not elem[-1]:
                 k = 2 if content_elem is not elem[0] else 1
-                self.parse_error("unexpected tag %r after complexContent declaration:" % elem[k].tag, elem)
+                self.parse_error(
+                    "unexpected tag %r after complexContent declaration:" % elem[k].tag, elem
+                )
 
         elif content_elem.tag == XSD_OPEN_CONTENT and self.xsd_version > '1.0':
             self.open_content = XsdOpenContent(content_elem, self.schema, self)
@@ -207,7 +207,9 @@ class XsdComplexType(XsdType, ValidationMixin):
                     if content_elem is not child:
                         continue
                     elif elem[index + 1].tag in {XSD_GROUP, XSD_SEQUENCE, XSD_ALL, XSD_CHOICE}:
-                        self.content_type = self.schema.BUILDERS.group_class(elem[index + 1], self.schema, self)
+                        self.content_type = self.schema.BUILDERS.group_class(
+                            elem[index + 1], self.schema, self
+                        )
                     else:
                         self.content_type = self.schema.self.schema.create_empty_content_group(self)
                     break
@@ -216,7 +218,9 @@ class XsdComplexType(XsdType, ValidationMixin):
         else:
             if self.schema.validation == 'skip':
                 # Also generated by meta-schema validation for 'lax' and 'strict' modes
-                self.parse_error("unexpected tag %r for complexType content:" % content_elem.tag, elem)
+                self.parse_error(
+                    "unexpected tag %r for complexType content:" % content_elem.tag, elem
+                )
             self.content_type = self.schema.create_any_content_group(self)
             self.attributes = self.schema.create_any_attribute_group(self)
 
@@ -227,7 +231,9 @@ class XsdComplexType(XsdType, ValidationMixin):
             self.parse_error("wrong redefinition without self-reference", elem)
 
     def _parse_content_tail(self, elem, **kwargs):
-        self.attributes = self.schema.BUILDERS.attribute_group_class(elem, self.schema, self, **kwargs)
+        self.attributes = self.schema.BUILDERS.attribute_group_class(
+            elem, self.schema, self, **kwargs
+        )
 
     def _parse_derivation_elem(self, elem):
         derivation_elem = self._parse_child_component(elem)
@@ -241,7 +247,9 @@ class XsdComplexType(XsdType, ValidationMixin):
         if self.derivation is None:
             self.derivation = derivation
         elif self.redefine is None:
-            raise XMLSchemaValueError("%r is expected to have a redefined/overridden component" % self)
+            raise XMLSchemaValueError(
+                "%r is expected to have a redefined/overridden component" % self
+            )
 
         if self.base_type is not None and derivation in self.base_type.final:
             self.parse_error("%r derivation not allowed for %r." % (derivation, self))
@@ -267,7 +275,8 @@ class XsdComplexType(XsdType, ValidationMixin):
                 return self.maps.types[XSD_ANY_SIMPLE_TYPE]
         else:
             if isinstance(base_type, tuple):
-                self.parse_error("circularity definition found between %r and %r" % (self, base_qname), elem)
+                self.parse_error("circularity definition found between %r "
+                                 "and %r" % (self, base_qname), elem)
                 return self.maps.types[XSD_ANY_TYPE]
             elif complex_content and base_type.is_simple():
                 self.parse_error("a complexType ancestor required: %r" % base_type, elem)
@@ -299,7 +308,8 @@ class XsdComplexType(XsdType, ValidationMixin):
                                  "an element-only content type ", base_type.elem)
                 self.content_type = self.schema.create_any_content_group(self)
 
-            self._parse_content_tail(elem, derivation='restriction', base_attributes=base_type.attributes)
+            self._parse_content_tail(elem, derivation='restriction',
+                                     base_attributes=base_type.attributes)
 
     def _parse_simple_content_extension(self, elem, base_type):
         # simpleContent extension: the base type must be a simpleType or a complexType
@@ -318,7 +328,8 @@ class XsdComplexType(XsdType, ValidationMixin):
                 self.parse_error("base type %r has not simple content." % base_type, elem)
                 self.content_type = self.schema.create_any_content_group(self)
 
-            self._parse_content_tail(elem, derivation='extension', base_attributes=base_type.attributes)
+            self._parse_content_tail(elem, derivation='extension',
+                                     base_attributes=base_type.attributes)
 
     def _parse_complex_content_restriction(self, elem, base_type):
         if 'restriction' in base_type.final:
@@ -328,18 +339,22 @@ class XsdComplexType(XsdType, ValidationMixin):
             base_type = self.maps.types[XSD_ANY_TYPE]
 
         # complexContent restriction: the base type must be a complexType with a complex content.
-        for child in filter(lambda x: x.tag != XSD_ANNOTATION, elem):
+        for child in filter(is_not_xsd_annotation, elem):
             if child.tag == XSD_OPEN_CONTENT and self.xsd_version > '1.0':
                 self.open_content = XsdOpenContent(child, self.schema, self)
                 continue
             elif child.tag in XSD_MODEL_GROUP_TAGS:
                 content_type = self.schema.BUILDERS.group_class(child, self.schema, self)
                 if not base_type.content_type.admits_restriction(content_type.model):
-                    msg = "restriction of an xs:{} with more than one particle with xs:{} is forbidden"
-                    self.parse_error(msg.format(base_type.content_type.model, content_type.model))
+                    self.parse_error(
+                        "restriction of an xs:{} with more than one particle with xs:{} is "
+                        "forbidden".format(base_type.content_type.model, content_type.model)
+                    )
                 break
         else:
-            content_type = self.schema.create_empty_content_group(self, base_type.content_type.model)
+            content_type = self.schema.create_empty_content_group(
+                self, base_type.content_type.model
+            )
 
         content_type.restriction = base_type.content_type
 
@@ -364,13 +379,14 @@ class XsdComplexType(XsdType, ValidationMixin):
             self.parse_error(msg.format(self.open_content, base_type.open_content))
 
         self.content_type = content_type
-        self._parse_content_tail(elem, derivation='restriction', base_attributes=base_type.attributes)
+        self._parse_content_tail(elem, derivation='restriction',
+                                 base_attributes=base_type.attributes)
 
     def _parse_complex_content_extension(self, elem, base_type):
         if 'extension' in base_type.final:
             self.parse_error("the base type is not derivable by extension")
 
-        for group_elem in filter(lambda x: x.tag != XSD_ANNOTATION, elem):
+        for group_elem in filter(is_not_xsd_annotation, elem):
             break
         else:
             group_elem = None
@@ -379,7 +395,9 @@ class XsdComplexType(XsdType, ValidationMixin):
             if not base_type.mixed:
                 # Empty element-only model extension: don't create a nested group.
                 if group_elem is not None and group_elem.tag in XSD_MODEL_GROUP_TAGS:
-                    self.content_type = self.schema.BUILDERS.group_class(group_elem, self.schema, self)
+                    self.content_type = self.schema.BUILDERS.group_class(
+                        group_elem, self.schema, self
+                    )
                 elif base_type.is_simple() or base_type.has_simple_content():
                     self.content_type = self.schema.create_empty_content_group(self)
                 else:
@@ -392,7 +410,9 @@ class XsdComplexType(XsdType, ValidationMixin):
                 self.content_type.append(self.schema.create_empty_content_group(self.content_type))
 
                 if group_elem is not None and group_elem.tag in XSD_MODEL_GROUP_TAGS:
-                    group = self.schema.BUILDERS.group_class(group_elem, self.schema, self.content_type)
+                    group = self.schema.BUILDERS.group_class(
+                        group_elem, self.schema, self.content_type
+                    )
                     if not self.mixed:
                         self.parse_error("base has a different content type (mixed=%r) and the "
                                          "extension group is not empty." % base_type.mixed, elem)
@@ -423,7 +443,9 @@ class XsdComplexType(XsdType, ValidationMixin):
             content_type.elem.append(group.elem)
 
             if base_type.content_type.model == 'all' and base_type.content_type and group:
-                self.parse_error("XSD 1.0 does not allow extension of a not empty 'all' model group")
+                self.parse_error(
+                    "XSD 1.0 does not allow extension of a not empty 'all' model group"
+                )
             if base_type.mixed != self.mixed and base_type.name != XSD_ANY_TYPE:
                 self.parse_error("base has a different content type (mixed=%r) and the "
                                  "extension group is not empty." % base_type.mixed, elem)
@@ -434,7 +456,9 @@ class XsdComplexType(XsdType, ValidationMixin):
             self.content_type.append(base_type.content_type)
             self.content_type.elem.append(base_type.content_type.elem)
             if base_type.mixed != self.mixed and base_type.name != XSD_ANY_TYPE and self.mixed:
-                self.parse_error("extended type has a mixed content but the base is element-only", elem)
+                self.parse_error(
+                    "extended type has a mixed content but the base is element-only", elem
+                )
         else:
             self.content_type = self.schema.create_empty_content_group(self)
 
@@ -486,8 +510,6 @@ class XsdComplexType(XsdType, ValidationMixin):
             return False
 
     def is_element_only(self):
-        if self.name == XSD_ANY_TYPE:
-            return False
         try:
             return not self.content_type.mixed
         except AttributeError:
@@ -522,7 +544,8 @@ class XsdComplexType(XsdType, ValidationMixin):
                 return False
             return self.content_type.is_derived(other, derivation)
         elif self.has_simple_content():
-            return self.content_type.is_derived(other, derivation) or self.base_type.is_derived(other, derivation)
+            return self.content_type.is_derived(other, derivation) or \
+                self.base_type.is_derived(other, derivation)
         else:
             return self.base_type.is_derived(other, derivation)
 
@@ -530,14 +553,11 @@ class XsdComplexType(XsdType, ValidationMixin):
         if xsd_classes is None or isinstance(self, xsd_classes):
             yield self
         if self.attributes.parent is not None:
-            for obj in self.attributes.iter_components(xsd_classes):
-                yield obj
+            yield from self.attributes.iter_components(xsd_classes)
         if self.content_type.parent is not None:
-            for obj in self.content_type.iter_components(xsd_classes):
-                yield obj
+            yield from self.content_type.iter_components(xsd_classes)
         if getattr(self.base_type, 'parent', None) is not None:
-            for obj in self.base_type.iter_components(xsd_classes):
-                yield obj
+            yield from self.base_type.iter_components(xsd_classes)
 
         for obj in filter(lambda x: x.base_type is self, self.assertions):
             if xsd_classes is None or isinstance(obj, xsd_classes):
@@ -580,15 +600,14 @@ class XsdComplexType(XsdType, ValidationMixin):
         element is not available.
 
         :param elem: the Element that has to be decoded.
-        :param validation: the validation mode. Can be 'lax', 'strict' or 'skip.
+        :param validation: the validation mode. Can be 'lax', 'strict' or 'skip'.
         :param kwargs: keyword arguments for the decoding process.
         :return: yields a decoded object, eventually preceded by a sequence of \
         validation or decoding errors.
         """
         xsd_element = self.schema.create_element(name=elem.tag)
         xsd_element.type = self
-        for result in xsd_element.iter_decode(elem, validation, **kwargs):
-            yield result
+        yield from xsd_element.iter_decode(elem, validation, **kwargs)
 
     def iter_encode(self, obj, validation='lax', **kwargs):
         """
@@ -614,12 +633,10 @@ class XsdComplexType(XsdType, ValidationMixin):
             except XMLSchemaValueError:
                 pass
             else:
-                for result in results:
-                    yield result
+                yield from results
                 return
 
-        for result in xsd_element.iter_encode(value, validation, **kwargs):
-            yield result
+        yield from xsd_element.iter_encode(value, validation, **kwargs)
 
 
 class Xsd11ComplexType(XsdComplexType):
@@ -636,7 +653,8 @@ class Xsd11ComplexType(XsdComplexType):
           defaultAttributesApply = boolean : true
           {any attributes with non-schema namespace . . .}>
           Content: (annotation?, (simpleContent | complexContent | (openContent?,
-          (group | all | choice | sequence)?, ((attribute | attributeGroup)*, anyAttribute?), assert*)))
+          (group | all | choice | sequence)?,
+          ((attribute | attributeGroup)*, anyAttribute?), assert*)))
         </complexType>
     """
     default_attributes_apply = True
@@ -648,7 +666,7 @@ class Xsd11ComplexType(XsdComplexType):
         if self.redefine is not None:
             return self.schema.default_attributes
 
-        for child in filter(lambda x: x.tag == XSD_OVERRIDE, self.schema.root):
+        for child in filter(is_xsd_override, self.schema.root):
             if self.elem in child:
                 schema = self.schema.includes[child.attrib['schemaLocation']]
                 if schema.override is self.schema:
@@ -661,7 +679,7 @@ class Xsd11ComplexType(XsdComplexType):
         if self.parent is not None:
             return self.schema.default_open_content
 
-        for child in filter(lambda x: x.tag == XSD_OVERRIDE, self.schema.root):
+        for child in filter(is_xsd_override, self.schema.root):
             if self.elem in child:
                 schema = self.schema.includes[child.attrib['schemaLocation']]
                 if schema.override is self.schema:
@@ -684,7 +702,8 @@ class Xsd11ComplexType(XsdComplexType):
             if self.open_content is None:
                 assert self.content_type.interleave is None and self.content_type.suffix is None
             elif self.open_content.mode == 'interleave':
-                self.content_type.interleave = self.content_type.suffix = self.open_content.any_element
+                self.content_type.interleave = self.content_type.suffix \
+                    = self.open_content.any_element
             elif self.open_content.mode == 'suffix':
                 self.content_type.suffix = self.open_content.any_element
 
@@ -706,7 +725,9 @@ class Xsd11ComplexType(XsdComplexType):
         # Add default attributes
         if self.default_attributes_apply and isinstance(self.default_attributes, XsdComponent):
             if self.redefine is None and any(k in self.attributes for k in self.default_attributes):
-                self.parse_error("at least a default attribute is already declared in the complex type")
+                self.parse_error(
+                    "at least a default attribute is already declared in the complex type"
+                )
             self.attributes.update((k, v) for k, v in self.default_attributes.items())
 
     def _parse_complex_content_extension(self, elem, base_type):
@@ -721,7 +742,7 @@ class Xsd11ComplexType(XsdComplexType):
             self.parse_error("the base type is not derivable by extension")
 
         # Parse openContent
-        for group_elem in filter(lambda x: x.tag != XSD_ANNOTATION, elem):
+        for group_elem in filter(is_not_xsd_annotation, elem):
             if group_elem.tag != XSD_OPEN_CONTENT:
                 break
             self.open_content = XsdOpenContent(group_elem, self.schema, self)
@@ -736,7 +757,9 @@ class Xsd11ComplexType(XsdComplexType):
             if not base_type.mixed:
                 # Empty element-only model extension: don't create a nested sequence group.
                 if group_elem is not None and group_elem.tag in XSD_MODEL_GROUP_TAGS:
-                    self.content_type = self.schema.BUILDERS.group_class(group_elem, self.schema, self)
+                    self.content_type = self.schema.BUILDERS.group_class(
+                        group_elem, self.schema, self
+                    )
                 else:
                     self.content_type = self.schema.create_empty_content_group(
                         parent=self, model=base_type.content_type.model
@@ -747,7 +770,9 @@ class Xsd11ComplexType(XsdComplexType):
                 self.content_type.append(self.schema.create_empty_content_group(self.content_type))
 
                 if group_elem is not None and group_elem.tag in XSD_MODEL_GROUP_TAGS:
-                    group = self.schema.BUILDERS.group_class(group_elem, self.schema, self.content_type)
+                    group = self.schema.BUILDERS.group_class(
+                        group_elem, self.schema, self.content_type
+                    )
                     if not self.mixed:
                         self.parse_error("base has a different content type (mixed=%r) and the "
                                          "extension group is not empty." % base_type.mixed, elem)
@@ -782,7 +807,9 @@ class Xsd11ComplexType(XsdComplexType):
                 if not group:
                     pass
                 elif group.model != 'all':
-                    self.parse_error("cannot extend a not empty 'all' model group with a different model")
+                    self.parse_error(
+                        "cannot extend a not empty 'all' model group with a different model"
+                    )
                 elif base_type.content_type.min_occurs != group.min_occurs:
                     self.parse_error("when extend an xs:all group minOccurs must be the same")
                 elif base_type.mixed and not base_type.content_type:
@@ -802,7 +829,9 @@ class Xsd11ComplexType(XsdComplexType):
             self.content_type.append(base_type.content_type)
             self.content_type.elem.append(base_type.content_type.elem)
             if base_type.mixed != self.mixed and base_type.name != XSD_ANY_TYPE and self.mixed:
-                self.parse_error("extended type has a mixed content but the base is element-only", elem)
+                self.parse_error(
+                    "extended type has a mixed content but the base is element-only", elem
+                )
         else:
             self.content_type = self.schema.create_empty_content_group(self)
 
@@ -824,8 +853,11 @@ class Xsd11ComplexType(XsdComplexType):
         self._parse_content_tail(elem, derivation='extension', base_attributes=base_type.attributes)
 
     def _parse_content_tail(self, elem, **kwargs):
-        self.attributes = self.schema.BUILDERS.attribute_group_class(elem, self.schema, self, **kwargs)
+        self.attributes = self.schema.BUILDERS.attribute_group_class(
+            elem, self.schema, self, **kwargs
+        )
 
-        self.assertions = [XsdAssert(e, self.schema, self, self) for e in elem if e.tag == XSD_ASSERT]
+        self.assertions = [XsdAssert(e, self.schema, self, self)
+                           for e in elem if e.tag == XSD_ASSERT]
         if getattr(self.base_type, 'assertions', None):
             self.assertions.extend(assertion for assertion in self.base_type.assertions)
