@@ -1,6 +1,5 @@
-# -*- coding: utf-8 -*-
 #
-# Copyright (c), 2016-2019, SISSA (International School for Advanced Studies).
+# Copyright (c), 2016-2020, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -11,26 +10,26 @@
 """
 This module contains classes for XML Schema elements, complex types and model groups.
 """
-from __future__ import unicode_literals
 import warnings
 from decimal import Decimal
 from elementpath import XPath2Parser, ElementPathError, XPathContext
 from elementpath.datatypes import AbstractDateTime, Duration
 
-from ..exceptions import XMLSchemaAttributeError
-from ..qnames import XSD_ANNOTATION, XSD_GROUP, XSD_SEQUENCE, XSD_ALL, \
-    XSD_CHOICE, XSD_ATTRIBUTE_GROUP, XSD_COMPLEX_TYPE, XSD_SIMPLE_TYPE, \
+from ..exceptions import XMLSchemaAttributeError, XMLSchemaValueError
+from ..qnames import XSD_GROUP, XSD_SEQUENCE, XSD_ALL, XSD_CHOICE, \
+    XSD_ATTRIBUTE_GROUP, XSD_COMPLEX_TYPE, XSD_SIMPLE_TYPE, \
     XSD_ALTERNATIVE, XSD_ELEMENT, XSD_ANY_TYPE, XSD_UNIQUE, XSD_KEY, \
-    XSD_KEYREF, XSI_NIL, XSI_TYPE, XSD_ERROR, get_qname
+    XSD_KEYREF, XSI_NIL, XSI_TYPE, XSD_ERROR, get_qname, is_not_xsd_annotation
 from ..etree import etree_element, etree_iter_location_hints
 from ..helpers import get_xsd_derivation_attribute, get_xsd_form_attribute, \
-    ParticleCounter, strictly_equal
+    raw_xml_encode, ParticleCounter, strictly_equal
 from ..namespaces import get_namespace
-from ..converters import ElementData, raw_xml_encode, XMLSchemaConverter
+from ..converters import ElementData, XMLSchemaConverter
 from ..xpath import XMLSchemaProxy, ElementPathMixin
 
 from .exceptions import XMLSchemaValidationError, XMLSchemaTypeTableWarning
-from .xsdbase import XsdComponent, XsdType, ValidationMixin, ParticleMixin
+from .xsdbase import XSD_TYPE_DERIVATIONS, XSD_ELEMENT_DERIVATIONS, \
+    XsdComponent, XsdType, ValidationMixin, ParticleMixin
 from .identities import XsdKeyref
 from .wildcards import XsdAnyElement
 
@@ -65,7 +64,6 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
           Content: (annotation?, ((simpleType | complexType)?, (unique | key | keyref)*))
         </element>
     """
-    type = None
     qualified = False
     alternatives = ()
     inheritable = ()
@@ -77,6 +75,7 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
     _form = None
     _nillable = False
     _substitution_group = None
+    _head_type = None
 
     def __init__(self, elem, schema, parent):
         super(XsdElement, self).__init__(elem, schema, parent)
@@ -88,9 +87,13 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
 
     def __repr__(self):
         if self.ref is None:
-            return '%s(name=%r, occurs=%r)' % (self.__class__.__name__, self.prefixed_name, self.occurs)
+            return '%s(name=%r, occurs=%r)' % (
+                self.__class__.__name__, self.prefixed_name, self.occurs
+            )
         else:
-            return '%s(ref=%r, occurs=%r)' % (self.__class__.__name__, self.prefixed_name, self.occurs)
+            return '%s(ref=%r, occurs=%r)' % (
+                self.__class__.__name__, self.prefixed_name, self.occurs
+            )
 
     def __setattr__(self, name, value):
         if name == "type":
@@ -103,8 +106,7 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
 
     def __iter__(self):
         if not self.type.has_simple_content():
-            for e in self.type.content_type.iter_elements():
-                yield e
+            yield from self.type.content_type.iter_elements()
 
     @property
     def xpath_proxy(self):
@@ -136,7 +138,8 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
             for attr_name in ('type', 'nillable', 'default', 'fixed', 'form',
                               'block', 'abstract', 'final', 'substitutionGroup'):
                 if attr_name in attrib:
-                    self.parse_error("attribute %r is not allowed when element reference is used." % attr_name)
+                    self.parse_error("attribute %r is not allowed when "
+                                     "element reference is used." % attr_name)
             return
 
         if 'form' in attrib:
@@ -168,7 +171,7 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
         if 'block' in attrib:
             try:
                 self._block = get_xsd_derivation_attribute(
-                    self.elem, 'block', ('extension', 'restriction', 'substitution')
+                    self.elem, 'block', XSD_ELEMENT_DERIVATIONS
                 )
             except ValueError as err:
                 self.parse_error(err)
@@ -179,17 +182,23 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
         if self.parent is None:
             if 'final' in attrib:
                 try:
-                    self._final = get_xsd_derivation_attribute(self.elem, 'final', ('extension', 'restriction'))
+                    self._final = get_xsd_derivation_attribute(
+                        self.elem, 'final', XSD_TYPE_DERIVATIONS
+                    )
                 except ValueError as err:
                     self.parse_error(err)
 
-            for attr_name in ('ref', 'form', 'minOccurs', 'maxOccurs'):
+            for attr_name in {'ref', 'form', 'minOccurs', 'maxOccurs'}:
                 if attr_name in attrib:
-                    self.parse_error("attribute %r not allowed in a global element declaration" % attr_name)
+                    self.parse_error(
+                        "attribute %r not allowed in a global element declaration" % attr_name
+                    )
         else:
-            for attr_name in ('final', 'substitutionGroup'):
+            for attr_name in {'final', 'substitutionGroup'}:
                 if attr_name in attrib:
-                    self.parse_error("attribute %r not allowed in a local element declaration" % attr_name)
+                    self.parse_error(
+                        "attribute %r not allowed in a local element declaration" % attr_name
+                    )
 
     def _parse_type(self):
         attrib = self.elem.attrib
@@ -211,8 +220,8 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
             finally:
                 child = self._parse_child_component(self.elem, strict=False)
                 if child is not None and child.tag in (XSD_COMPLEX_TYPE, XSD_SIMPLE_TYPE):
-                    msg = "the attribute 'type' and the <%s> local declaration are mutually exclusive"
-                    self.parse_error(msg % child.tag.split('}')[-1])
+                    self.parse_error("the attribute 'type' and the <%s> local declaration "
+                                     "are mutually exclusive" % child.tag.split('}')[-1])
         else:
             child = self._parse_child_component(self.elem, strict=False)
             if child is not None:
@@ -259,7 +268,7 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
             return
 
         self.identities = {}
-        for child in filter(lambda x: x.tag != XSD_ANNOTATION, self.elem[index:]):
+        for child in filter(is_not_xsd_annotation, self.elem[index:]):
             if child.tag == XSD_UNIQUE:
                 constraint = self.schema.BUILDERS.unique_class(child, self.schema, self)
             elif child.tag == XSD_KEY:
@@ -291,7 +300,9 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
             return
         else:
             if substitution_group_qname[0] != '{':
-                substitution_group_qname = get_qname(self.target_namespace, substitution_group_qname)
+                substitution_group_qname = get_qname(
+                    self.target_namespace, substitution_group_qname
+                )
 
         try:
             head_element = self.maps.lookup_element(substitution_group_qname)
@@ -306,20 +317,25 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
                 return
 
         final = head_element.final
-        if self.type == head_element.type or self.type.name == XSD_ANY_TYPE:
+        if self.type == head_element.type:
             pass
+        elif self.type.name == XSD_ANY_TYPE:
+            if head_element.type.name != XSD_ANY_TYPE:
+                # Use head element's type for validate content
+                # ref: https://www.w3.org/TR/xmlschema-1/#cElement_Declarations
+                self._head_type = head_element.type
         elif not self.type.is_derived(head_element.type):
-            msg = "%r type is not of the same or a derivation of the head element %r type."
-            self.parse_error(msg % (self, head_element))
+            self.parse_error("%r type is not of the same or a derivation "
+                             "of the head element %r type." % (self, head_element))
         elif final == '#all' or 'extension' in final and 'restriction' in final:
-            msg = "head element %r can't be substituted by an element that has a derivation of its type"
-            self.parse_error(msg % head_element)
+            self.parse_error("head element %r can't be substituted by an element "
+                             "that has a derivation of its type" % head_element)
         elif 'extension' in final and self.type.is_derived(head_element.type, 'extension'):
-            msg = "head element %r can't be substituted by an element that has an extension of its type"
-            self.parse_error(msg % head_element)
+            self.parse_error("head element %r can't be substituted by an element "
+                             "that has an extension of its type" % head_element)
         elif 'restriction' in final and self.type.is_derived(head_element.type, 'restriction'):
-            msg = "head element %r can't be substituted by an element that has a restriction of its type"
-            self.parse_error(msg % head_element)
+            self.parse_error("head element %r can't be substituted by an element "
+                             "that has a restriction of its type" % head_element)
 
         try:
             self.maps.substitution_groups[substitution_group_qname].add(self)
@@ -391,13 +407,16 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
         return self.type.attributes[name]
 
     def get_type(self, elem, inherited=None):
-        return self.type
+        return self._head_type or self.type
 
     def get_attributes(self, xsd_type):
         try:
             return xsd_type.attributes
         except AttributeError:
-            return self.attributes
+            if xsd_type is self.type:
+                return self.attributes
+            else:
+                return self.schema.create_empty_attribute_group(self)
 
     def get_path(self, ancestor=None, reverse=False):
         """
@@ -405,7 +424,8 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
         in which the element is contained or is relative to a specific ancestor passed as argument.
         In the latter case returns `None` if the argument is not an ancestor.
 
-        :param ancestor: optional XSD component of the same schema, that may be an ancestor of the element.
+        :param ancestor: optional XSD component of the same schema, that maybe \
+        an ancestor of the element.
         :param reverse: if set to `True` returns the reverse path, from the element to ancestor.
         """
         path = []
@@ -423,8 +443,7 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
     def iter_components(self, xsd_classes=None):
         if xsd_classes is None:
             yield self
-            for obj in self.identities.values():
-                yield obj
+            yield from self.identities.values()
         else:
             if isinstance(self, xsd_classes):
                 yield self
@@ -432,9 +451,11 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
                 if isinstance(obj, xsd_classes):
                     yield obj
 
+        if not hasattr(self.type, 'attributes'):
+            yield from self.attributes.iter_components(xsd_classes)
+
         if self.ref is None and self.type.parent is not None:
-            for obj in self.type.iter_components(xsd_classes):
-                yield obj
+            yield from self.type.iter_components(xsd_classes)
 
     def iter_substitutes(self):
         for xsd_element in self.maps.substitution_groups.get(self.name, ()):
@@ -449,6 +470,8 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
         text = elem.text
         if text is None:
             text = self.fixed if self.fixed is not None else self.default
+            if text is None:
+                return
         return self.type.text_decode(text)
 
     def check_dynamic_context(self, elem, **kwargs):
@@ -485,19 +508,43 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
             if ns not in locations:
                 locations[ns] = None
 
+    def start_identities(self, identities):
+        """
+        Start tracking of XSD element's identities.
+
+        :param identities: a dictionary containing the identities counters.
+        """
+        for constraint in self.identities.values():
+            try:
+                identities[constraint].clear()
+            except KeyError:
+                identities[constraint] = constraint.get_counter()
+
+    def stop_identities(self, identities):
+        """
+        Stop tracking of XSD element's identities.
+
+        :param identities: a dictionary containing the identities counters.
+        """
+        for identity in self.identities.values():
+            try:
+                identities[identity].enabled = False
+            except KeyError:
+                identities[identity] = identity.get_counter(enabled=False)
+
     def iter_decode(self, elem, validation='lax', **kwargs):
         """
         Creates an iterator for decoding an Element instance.
 
         :param elem: the Element that has to be decoded.
-        :param validation: the validation mode, can be 'lax', 'strict' or 'skip.
+        :param validation: the validation mode, can be 'lax', 'strict' or 'skip'.
         :param kwargs: keyword arguments for the decoding process.
         :return: yields a decoded object, eventually preceded by a sequence of \
         validation or decoding errors.
         """
         if self.abstract:
-            msg = "cannot use an abstract element for validation"
-            yield self.validation_error(validation, msg, elem, **kwargs)
+            reason = "cannot use an abstract element for validation"
+            yield self.validation_error(validation, reason, elem, **kwargs)
 
         try:
             namespaces = kwargs['namespaces']
@@ -509,6 +556,13 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
         except KeyError:
             level = kwargs['level'] = 0
 
+        try:
+            identities = kwargs['identities']
+        except KeyError:
+            identities = kwargs['identities'] = {}
+
+        self.start_identities(identities)
+
         try:
             converter = kwargs['converter']
         except KeyError:
@@ -518,7 +572,7 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
                 converter = kwargs['converter'] = self.schema.get_converter(**kwargs)
 
         try:
-            pass  # self.check_dynamic_context(elem, **kwargs)
+            pass  # self.check_dynamic_context(elem, **kwargs) TODO: dynamic schema load
         except XMLSchemaValidationError as err:
             yield self.validation_error(validation, err, elem, **kwargs)
 
@@ -535,7 +589,13 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
                 yield self.validation_error(validation, err, elem, **kwargs)
 
             if xsd_type.is_blocked(self):
-                yield self.validation_error(validation, "usage of %r is blocked" % xsd_type, elem, **kwargs)
+                reason = "usage of %r is blocked" % xsd_type
+                yield self.validation_error(validation, reason, elem, **kwargs)
+
+        if xsd_type.abstract:
+            yield self.validation_error(validation, "%r is abstract", elem, **kwargs)
+
+        content_type = xsd_type.content_type if xsd_type.is_complex() else xsd_type
 
         # Decode attributes
         attribute_group = self.get_attributes(xsd_type)
@@ -557,7 +617,8 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
         if XSI_NIL in elem.attrib:
             xsi_nil = elem.attrib[XSI_NIL].strip()
             if not self.nillable:
-                yield self.validation_error(validation, "element is not nillable.", elem, **kwargs)
+                reason = "element is not nillable."
+                yield self.validation_error(validation, reason, elem, **kwargs)
             elif xsi_nil not in {'0', '1', 'false', 'true'}:
                 reason = "xsi:nil attribute must has a boolean value."
                 yield self.validation_error(validation, reason, elem, **kwargs)
@@ -572,7 +633,7 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
             else:
                 if converter is not None:
                     element_data = ElementData(elem.tag, None, None, attributes)
-                    yield converter.element_decode(element_data, self, level)
+                    yield converter.element_decode(element_data, self, xsd_type, level)
                 return
 
         if xsd_type.is_empty() and elem.text:
@@ -584,7 +645,7 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
                 for error in assertion(elem, **kwargs):
                     yield self.validation_error(validation, error, **kwargs)
 
-            for result in xsd_type.content_type.iter_decode(elem, validation, **kwargs):
+            for result in content_type.iter_decode(elem, validation, **kwargs):
                 if isinstance(result, XMLSchemaValidationError):
                     yield self.validation_error(validation, result, elem, **kwargs)
                 else:
@@ -593,8 +654,13 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
             if len(content) == 1 and content[0][0] == 1:
                 value, content = content[0][1], None
 
+            if self.fixed is not None and \
+                    (len(elem) > 0 or value is not None and self.fixed != value):
+                reason = "must have the fixed value %r." % self.fixed
+                yield self.validation_error(validation, reason, elem, **kwargs)
+
         else:
-            if len(elem) and validation != 'skip':
+            if len(elem):
                 reason = "a simple content element can't has child elements."
                 yield self.validation_error(validation, reason, elem, **kwargs)
 
@@ -602,10 +668,11 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
             if self.fixed is not None:
                 if text is None:
                     text = self.fixed
-                elif text == self.fixed or validation == 'skip':
+                elif text == self.fixed:
                     pass
-                elif not strictly_equal(xsd_type.text_decode(text), xsd_type.text_decode(self.fixed)):
-                    reason = "must has the fixed value %r." % self.fixed
+                elif not strictly_equal(xsd_type.text_decode(text),
+                                        xsd_type.text_decode(self.fixed)):
+                    reason = "must have the fixed value %r." % self.fixed
                     yield self.validation_error(validation, reason, elem, **kwargs)
 
             elif not text and kwargs.get('use_defaults') and self.default is not None:
@@ -616,21 +683,21 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
                     for error in assertion(elem, value=text, **kwargs):
                         yield self.validation_error(validation, error, **kwargs)
 
-                if text and xsd_type.content_type.is_list():
+                if text and content_type.is_list():
                     value = text.split()
                 else:
                     value = text
-
-                xsd_type = xsd_type.content_type
+            elif xsd_type.is_key() and self.xsd_version != '1.0':
+                kwargs['element'] = self if self.ref is None else self.ref
 
             if text is None:
-                for result in xsd_type.iter_decode('', validation, **kwargs):
+                for result in content_type.iter_decode('', validation, **kwargs):
                     if isinstance(result, XMLSchemaValidationError):
                         yield self.validation_error(validation, result, elem, **kwargs)
                         if 'filler' in kwargs:
                             value = kwargs['filler'](self)
             else:
-                for result in xsd_type.iter_decode(text, validation, **kwargs):
+                for result in content_type.iter_decode(text, validation, **kwargs):
                     if isinstance(result, XMLSchemaValidationError):
                         yield self.validation_error(validation, result, elem, **kwargs)
                     elif result is None and 'filler' in kwargs:
@@ -638,34 +705,65 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
                     else:
                         value = result
 
-        if isinstance(value, Decimal):
-            try:
-                value = kwargs['decimal_type'](value)
-            except (KeyError, TypeError):
-                pass
-        elif isinstance(value, (AbstractDateTime, Duration)):
-            try:
-                if kwargs['datetime_types'] is not True:
+            if isinstance(value, Decimal):
+                try:
+                    value = kwargs['decimal_type'](value)
+                except (KeyError, TypeError):
+                    pass
+            elif isinstance(value, (AbstractDateTime, Duration)):
+                try:
+                    if kwargs['datetime_types'] is not True:
+                        value = elem.text
+                except KeyError:
                     value = elem.text
-            except KeyError:
-                value = elem.text
+            elif isinstance(value, str) and value.startswith('{'):
+                if xsd_type.is_qname():
+                    value = text
 
         if converter is not None:
             element_data = ElementData(elem.tag, value, content, attributes)
-            yield converter.element_decode(element_data, self, level)
+            yield converter.element_decode(element_data, self, xsd_type, level)
+        elif not level:
+            yield ElementData(elem.tag, value, None, attributes)
+
         if content is not None:
             del content
 
-        if validation != 'skip':
-            if 'max_depth' in kwargs:
-                # Don't check key references with lazy or shallow validation
-                for constraint in filter(lambda x: not isinstance(x, XsdKeyref), self.identities.values()):
-                    for error in constraint(elem, namespaces):
-                        yield self.validation_error(validation, error, elem, **kwargs)
+        # Collects fields values for identities that refer to this element.
+        for constraint, counter in identities.items():
+            if not counter.enabled or self not in constraint.elements:
+                continue
+
+            try:
+                if xsd_type is self.type:
+                    xsd_fields = constraint.get_fields(self)
+                else:
+                    xsd_element = self.copy()
+                    xsd_element.type = xsd_type
+                    xsd_fields = constraint.get_fields(xsd_element)
+
+                if not any(fld is not None for fld in xsd_fields):
+                    continue
+                fields = constraint.get_fields(elem, namespaces, decoders=xsd_fields)
+            except XMLSchemaValueError as err:
+                yield self.validation_error(validation, err, elem, **kwargs)
             else:
-                for constraint in self.identities.values():
-                    for error in constraint(elem, namespaces):
-                        yield self.validation_error(validation, error, elem, **kwargs)
+                if any(fld is not None for fld in fields):
+                    try:
+                        counter.increase(fields)
+                    except ValueError as err:
+                        yield self.validation_error(validation, err, elem, **kwargs)
+
+        # Disable collect for out of scope identities and check key references
+        if 'max_depth' not in kwargs:
+            for constraint in self.identities.values():
+                counter = identities[constraint]
+                counter.enabled = False
+                if isinstance(constraint, XsdKeyref):
+                    for err in counter.iter_errors(identities):
+                        yield self.validation_error(validation, err, elem, **kwargs)
+        elif level:
+            self.stop_identities(identities)
 
     def iter_encode(self, obj, validation='lax', **kwargs):
         """
@@ -746,7 +844,9 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
 
         elif xsd_type.has_simple_content():
             if element_data.text is not None:
-                for result in xsd_type.content_type.iter_encode(element_data.text, validation, **kwargs):
+                for result in xsd_type.content_type.iter_encode(
+                        element_data.text, validation, **kwargs
+                ):
                     if isinstance(result, XMLSchemaValidationError):
                         errors.append(result)
                     else:
@@ -760,7 +860,7 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
 
         elem = converter.etree_element(tag, text, children, attributes, level)
 
-        if validation != 'skip' and errors:
+        if errors:
             for e in errors:
                 yield self.validation_error(validation, e, elem, **kwargs)
         yield elem
@@ -825,7 +925,8 @@ class XsdElement(XsdComponent, ValidationMixin, ParticleMixin, ElementPathMixin)
             elif not self.is_consistent(other) and self.type.elem is not other.type.elem and \
                     not self.type.is_derived(other.type, 'restriction') and not other.type.abstract:
                 return False
-            elif self.fixed != other.fixed and self.type.normalize(self.fixed) != other.type.normalize(other.fixed):
+            elif self.fixed != other.fixed and \
+                    self.type.normalize(self.fixed) != other.type.normalize(other.fixed):
                 return False
             elif other.nillable is False and self.nillable:
                 return False
@@ -910,7 +1011,8 @@ class Xsd11Element(XsdElement):
           targetNamespace = anyURI
           type = QName
           {any attributes with non-schema namespace . . .}>
-          Content: (annotation?, ((simpleType | complexType)?, alternative*, (unique | key | keyref)*))
+          Content: (annotation?, ((simpleType | complexType)?, alternative*,
+          (unique | key | keyref)*))
         </element>
     """
     _target_namespace = None
@@ -937,7 +1039,7 @@ class Xsd11Element(XsdElement):
         else:
             alternatives = []
             has_test = True
-            for child in filter(lambda x: x.tag != XSD_ANNOTATION, self.elem[index:]):
+            for child in filter(is_not_xsd_annotation, self.elem[index:]):
                 if child.tag == XSD_ALTERNATIVE:
                     alternatives.append(XsdAlternative(child, self.schema, self))
                     if not has_test:
@@ -969,8 +1071,7 @@ class Xsd11Element(XsdElement):
     def iter_components(self, xsd_classes=None):
         if xsd_classes is None:
             yield self
-            for obj in self.identities.values():
-                yield obj
+            yield from self.identities.values()
         else:
             if isinstance(self, xsd_classes):
                 yield self
@@ -979,22 +1080,22 @@ class Xsd11Element(XsdElement):
                     yield obj
 
         for alt in self.alternatives:
-            for obj in alt.iter_components(xsd_classes):
-                yield obj
+            yield from alt.iter_components(xsd_classes)
+
+        if not hasattr(self.type, 'attributes'):
+            yield from self.attributes.iter_components(xsd_classes)
 
         if self.ref is None and self.type.parent is not None:
-            for obj in self.type.iter_components(xsd_classes):
-                yield obj
+            yield from self.type.iter_components(xsd_classes)
 
     def iter_substitutes(self):
         for xsd_element in self.maps.substitution_groups.get(self.name, ()):
             yield xsd_element
-            for e in xsd_element.iter_substitutes():
-                yield e
+            yield from xsd_element.iter_substitutes()
 
     def get_type(self, elem, inherited=None):
         if not self.alternatives:
-            return self.type
+            return self._head_type or self.type
 
         if isinstance(elem, ElementData):
             if elem.attributes:
@@ -1015,7 +1116,7 @@ class Xsd11Element(XsdElement):
                 if alt.token is None or alt.test(elem):
                     return alt.type
 
-        return self.type
+        return self._head_type or self.type
 
     def is_overlap(self, other):
         if isinstance(other, XsdElement):
@@ -1089,7 +1190,9 @@ class XsdAlternative(XsdComponent):
         super(XsdAlternative, self).__init__(elem, schema, parent)
 
     def __repr__(self):
-        return '%s(type=%r, test=%r)' % (self.__class__.__name__, self.elem.get('type'), self.elem.get('test'))
+        return '%s(type=%r, test=%r)' % (
+            self.__class__.__name__, self.elem.get('type'), self.elem.get('test')
+        )
 
     def __eq__(self, other):
         return self.path == other.path and self.type is other.type and \
@@ -1150,8 +1253,8 @@ class XsdAlternative(XsdComponent):
 
                 child = self._parse_child_component(self.elem, strict=False)
                 if child is not None and child.tag in (XSD_COMPLEX_TYPE, XSD_SIMPLE_TYPE):
-                    msg = "the attribute 'type' and the <%s> local declaration are mutually exclusive"
-                    self.parse_error(msg % child.tag.split('}')[-1])
+                    self.parse_error("the attribute 'type' and the <%s> local declaration "
+                                     "are mutually exclusive" % child.tag.split('}')[-1])
 
     @property
     def built(self):
@@ -1165,8 +1268,7 @@ class XsdAlternative(XsdComponent):
         if xsd_classes is None or isinstance(self, xsd_classes):
             yield self
         if self.type is not None and self.type.parent is not None:
-            for obj in self.type.iter_components(xsd_classes):
-                yield obj
+            yield from self.type.iter_components(xsd_classes)
 
     def test(self, elem):
         try:
