@@ -1,6 +1,5 @@
-# -*- coding: utf-8 -*-
 #
-# Copyright (c), 2016-2019, SISSA (International School for Advanced Studies).
+# Copyright (c), 2016-2020, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -11,9 +10,6 @@
 """
 This module contains classes for XML Schema wildcards.
 """
-from __future__ import unicode_literals
-
-from ..compat import unicode_type
 from ..exceptions import XMLSchemaValueError
 from ..namespaces import XSI_NAMESPACE
 from ..qnames import XSD_ANY, XSD_ANY_ATTRIBUTE, XSD_OPEN_CONTENT, \
@@ -70,7 +66,8 @@ class XsdWildcard(XsdComponent, ValidationMixin):
         if process_contents == 'strict':
             pass
         elif process_contents not in ('lax', 'skip'):
-            self.parse_error("wrong value %r for 'processContents' attribute" % self.process_contents)
+            self.parse_error("wrong value %r for 'processContents' "
+                             "attribute" % self.process_contents)
         else:
             self.process_contents = process_contents
 
@@ -99,8 +96,10 @@ class XsdWildcard(XsdComponent, ValidationMixin):
         not_qname = self.elem.attrib['notQName'].strip().split()
 
         if isinstance(self, XsdAnyAttribute) and \
-                not all(not s.startswith('##') or s == '##defined' for s in not_qname) or \
-                not all(not s.startswith('##') or s in {'##defined', '##definedSibling'} for s in not_qname):
+                not all(not s.startswith('##') or s == '##defined'
+                        for s in not_qname) or \
+                not all(not s.startswith('##') or s in {'##defined', '##definedSibling'}
+                        for s in not_qname):
             self.parse_error("wrong value for 'notQName' attribute")
             return
 
@@ -116,9 +115,12 @@ class XsdWildcard(XsdComponent, ValidationMixin):
 
         if self.not_namespace:
             if any(not x.startswith('##') for x in names) and \
-                    all(get_namespace(x) in self.not_namespace for x in names if not x.startswith('##')):
-                self.parse_error("the namespace of each QName in notQName is allowed by notNamespace")
-        elif any(not self.is_namespace_allowed(get_namespace(x)) for x in names if not x.startswith('##')):
+                    all(get_namespace(x) in self.not_namespace
+                        for x in names if not x.startswith('##')):
+                self.parse_error("the namespace of each QName in notQName "
+                                 "is allowed by notNamespace")
+        elif any(not self.is_namespace_allowed(get_namespace(x))
+                 for x in names if not x.startswith('##')):
             self.parse_error("names in notQName must be in namespaces that are allowed")
 
         self.not_qname = names
@@ -160,16 +162,20 @@ class XsdWildcard(XsdComponent, ValidationMixin):
 
     def deny_qnames(self, names):
         if self.not_namespace:
-            return all(x in self.not_qname or get_namespace(x) in self.not_namespace for x in names)
+            return all(x in self.not_qname or get_namespace(x) in self.not_namespace
+                       for x in names)
         elif '##any' in self.namespace:
             return all(x in self.not_qname for x in names)
         elif '##other' in self.namespace:
-            return all(x in self.not_qname or get_namespace(x) == self.target_namespace for x in names)
+            return all(x in self.not_qname or get_namespace(x) == self.target_namespace
+                       for x in names)
         else:
-            return all(x in self.not_qname or get_namespace(x) not in self.namespace for x in names)
+            return all(x in self.not_qname or get_namespace(x) not in self.namespace
+                       for x in names)
 
     def is_restriction(self, other, check_occurs=True):
-        if check_occurs and isinstance(self, ParticleMixin) and not self.has_occurs_restriction(other):
+        if check_occurs and isinstance(self, ParticleMixin) \
+                and not self.has_occurs_restriction(other):
             return False
         elif not isinstance(other, type(self)):
             return False
@@ -295,7 +301,8 @@ class XsdWildcard(XsdComponent, ValidationMixin):
 
         if self.not_namespace:
             if other.not_namespace:
-                self.not_namespace.extend(ns for ns in other.not_namespace if ns not in self.not_namespace)
+                self.not_namespace.extend(ns for ns in other.not_namespace
+                                          if ns not in self.not_namespace)
             elif '##any' in other.namespace:
                 pass
             elif '##other' not in other.namespace:
@@ -352,9 +359,9 @@ class XsdAnyElement(XsdWildcard, ParticleMixin, ElementPathMixin):
 
     ..  <any
           id = ID
-          maxOccurs = (nonNegativeInteger | unbounded)  : 1
+          maxOccurs = (nonNegativeInteger | unbounded) : 1
           minOccurs = nonNegativeInteger : 1
-          namespace = ((##any | ##other) | List of (anyURI | (##targetNamespace | ##local)) )  : ##any
+          namespace = ((##any | ##other) | List of (anyURI | (##targetNamespace|##local)) ) : ##any
           processContents = (lax | skip | strict) : strict
           {any attributes with non-schema namespace . . .}>
           Content: (annotation?)
@@ -425,9 +432,8 @@ class XsdAnyElement(XsdWildcard, ParticleMixin, ElementPathMixin):
 
     def iter_decode(self, elem, validation='lax', **kwargs):
         if not self.is_matching(elem.tag):
-            if validation != 'skip':
-                reason = "{!r} is not allowed here".format(elem)
-                yield self.validation_error(validation, reason, elem, **kwargs)
+            reason = "{!r} is not allowed here".format(elem)
+            yield self.validation_error(validation, reason, elem, **kwargs)
 
         elif self.process_contents == 'skip':
             return
@@ -437,18 +443,18 @@ class XsdAnyElement(XsdWildcard, ParticleMixin, ElementPathMixin):
                 xsd_element = self.maps.lookup_element(elem.tag)
             except LookupError:
                 if XSI_TYPE in elem.attrib:
-                    xsd_element = self.schema.create_element(name=elem.tag)
-                    for result in xsd_element.iter_decode(elem, validation, **kwargs):
-                        yield result
+                    if self.process_contents == 'lax':
+                        xsd_element = self.schema.create_element(name=elem.tag, nillable='true')
+                    else:
+                        xsd_element = self.schema.create_element(name=elem.tag)
+                    yield from xsd_element.iter_decode(elem, validation, **kwargs)
                 elif validation == 'skip' or self.process_contents == 'lax':
-                    for result in self.any_type.iter_decode(elem, validation, **kwargs):
-                        yield result
+                    yield from self.any_type.iter_decode(elem, validation, **kwargs)
                 else:
                     reason = "element %r not found." % elem.tag
                     yield self.validation_error(validation, reason, elem, **kwargs)
             else:
-                for result in xsd_element.iter_decode(elem, validation, **kwargs):
-                    yield result
+                yield from xsd_element.iter_decode(elem, validation, **kwargs)
 
         elif validation == 'skip':
             yield self.any_type.decode(elem) if len(elem) > 0 else elem.text
@@ -462,9 +468,8 @@ class XsdAnyElement(XsdWildcard, ParticleMixin, ElementPathMixin):
         namespace = get_namespace(name)
 
         if not self.is_namespace_allowed(namespace):
-            if validation != 'skip':
-                reason = "element {!r} is not allowed here".format(name)
-                yield self.validation_error(validation, reason, value, **kwargs)
+            reason = "element {!r} is not allowed here".format(name)
+            yield self.validation_error(validation, reason, value, **kwargs)
 
         elif self.process_contents == 'skip':
             return
@@ -474,14 +479,12 @@ class XsdAnyElement(XsdWildcard, ParticleMixin, ElementPathMixin):
                 xsd_element = self.maps.lookup_element(name)
             except LookupError:
                 if validation == 'skip' or self.process_contents == 'lax':
-                    for result in self.any_type.iter_encode(obj, validation, **kwargs):
-                        yield result
+                    yield from self.any_type.iter_encode(obj, validation, **kwargs)
                 elif self.process_contents == 'strict':
                     reason = "element %r not found." % name
                     yield self.validation_error(validation, reason, **kwargs)
             else:
-                for result in xsd_element.iter_encode(value, validation, **kwargs):
-                    yield result
+                yield from xsd_element.iter_encode(value, validation, **kwargs)
 
         elif validation == 'skip':
             yield self.any_type.encode(value)
@@ -567,9 +570,8 @@ class XsdAnyAttribute(XsdWildcard):
         name, value = attribute
 
         if not self.is_matching(name):
-            if validation != 'skip':
-                reason = "attribute %r not allowed." % name
-                yield self.validation_error(validation, reason, attribute, **kwargs)
+            reason = "attribute %r not allowed." % name
+            yield self.validation_error(validation, reason, attribute, **kwargs)
 
         elif self.process_contents == 'skip':
             return
@@ -584,8 +586,7 @@ class XsdAnyAttribute(XsdWildcard):
                     reason = "attribute %r not found." % name
                     yield self.validation_error(validation, reason, attribute, **kwargs)
             else:
-                for result in xsd_attribute.iter_decode(value, validation, **kwargs):
-                    yield result
+                yield from xsd_attribute.iter_decode(value, validation, **kwargs)
 
         elif validation == 'skip':
             yield value
@@ -599,9 +600,8 @@ class XsdAnyAttribute(XsdWildcard):
         namespace = get_namespace(name)
 
         if not self.is_namespace_allowed(namespace):
-            if validation != 'skip':
-                reason = "attribute %r not allowed." % name
-                yield self.validation_error(validation, reason, attribute, **kwargs)
+            reason = "attribute %r not allowed." % name
+            yield self.validation_error(validation, reason, attribute, **kwargs)
 
         elif self.process_contents == 'skip':
             return
@@ -611,16 +611,15 @@ class XsdAnyAttribute(XsdWildcard):
                 xsd_attribute = self.maps.lookup_attribute(name)
             except LookupError:
                 if validation == 'skip':
-                    yield unicode_type(value)
+                    yield str(value)
                 elif self.process_contents == 'strict':
                     reason = "attribute %r not found." % name
                     yield self.validation_error(validation, reason, attribute, **kwargs)
             else:
-                for result in xsd_attribute.iter_encode(value, validation, **kwargs):
-                    yield result
+                yield from xsd_attribute.iter_encode(value, validation, **kwargs)
 
         elif validation == 'skip':
-            yield unicode_type(value)
+            yield str(value)
 
         elif self.process_contents == 'strict':
             reason = "unavailable namespace {!r}".format(get_namespace(name))
@@ -669,14 +668,16 @@ class Xsd11AnyElement(XsdAnyElement):
                 return False
         else:
             name = '{%s}%s' % (default_namespace, name)
-            if not self.is_namespace_allowed('') and not self.is_namespace_allowed(default_namespace):
+            if not self.is_namespace_allowed('') \
+                    and not self.is_namespace_allowed(default_namespace):
                 return False
 
         if group in self.precedences:
             if occurs is None:
                 if any(e.is_matching(name) for e in self.precedences[group]):
                     return False
-            elif any(e.is_matching(name) and not e.is_over(occurs[e]) for e in self.precedences[group]):
+            elif any(e.is_matching(name) and not e.is_over(occurs[e])
+                     for e in self.precedences[group]):
                 return False
 
         if '##defined' in self.not_qname and name in self.maps.elements:
@@ -773,7 +774,8 @@ class XsdOpenContent(XsdComponent):
         child = self._parse_child_component(self.elem)
         if self.mode == 'none':
             if child is not None and child.tag == XSD_ANY:
-                self.parse_error("an openContent with mode='none' must not has an <xs:any> child declaration")
+                self.parse_error("an openContent with mode='none' must not "
+                                 "have an <xs:any> child declaration")
         elif child is None or child.tag != XSD_ANY:
             self.parse_error("an <xs:any> child declaration is required")
         else:
