@@ -1,6 +1,5 @@
-# -*- coding: utf-8 -*-
 #
-# Copyright (c), 2016-2019, SISSA (International School for Advanced Studies).
+# Copyright (c), 2016-2020, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -11,19 +10,18 @@
 """
 This module contains base functions and classes XML Schema components.
 """
-from __future__ import unicode_literals
 import re
 
-from ..compat import PY3, string_base_type, unicode_type
 from ..exceptions import XMLSchemaValueError, XMLSchemaTypeError
 from ..qnames import XSD_ANNOTATION, XSD_APPINFO, XSD_DOCUMENTATION, XML_LANG, \
-    XSD_ANY_TYPE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE, XSD_ID, XSD_OVERRIDE, \
-    get_qname, local_name, qname_to_prefixed
+    XSD_ANY_TYPE, XSD_ANY_SIMPLE_TYPE, XSD_ANY_ATOMIC_TYPE, XSD_ID, XSD_QNAME, \
+    XSD_OVERRIDE, get_qname, local_name, qname_to_prefixed, is_not_xsd_annotation
 from ..etree import etree_tostring
 from ..helpers import is_etree_element
-from .exceptions import XMLSchemaParseError, XMLSchemaValidationError, \
-    XMLSchemaDecodeError, XMLSchemaEncodeError
+from .exceptions import XMLSchemaParseError, XMLSchemaValidationError
 
+XSD_TYPE_DERIVATIONS = {'extension', 'restriction'}
+XSD_ELEMENT_DERIVATIONS = {'extension', 'restriction', 'substitution'}
 
 XSD_VALIDATION_MODES = {'strict', 'lax', 'skip'}
 """
@@ -32,6 +30,12 @@ Ref.: https://www.w3.org/TR/xmlschema11-1/#key-va
 """
 
 
+def check_validation_mode(validation):
+    if validation not in XSD_VALIDATION_MODES:
+        raise XMLSchemaValueError("validation mode can be 'strict', "
+                                  "'lax' or 'skip': %r" % validation)
+
+
 class XsdValidator(object):
     """
     Common base class for XML Schema validator, that represents a PSVI (Post Schema Validation
@@ -50,21 +54,12 @@ class XsdValidator(object):
     xsd_version = None
 
     def __init__(self, validation='strict'):
-        if validation not in XSD_VALIDATION_MODES:
-            raise XMLSchemaValueError("validation argument can be 'strict', 'lax' or 'skip': %r" % validation)
         self.validation = validation
         self.errors = []
 
     def __str__(self):
-        # noinspection PyCompatibility,PyUnresolvedReferences
-        return unicode(self).encode("utf-8")
-
-    def __unicode__(self):
         return self.__repr__()
 
-    if PY3:
-        __str__ = __unicode__
-
     @property
     def built(self):
         """
@@ -72,27 +67,31 @@ class XsdValidator(object):
         ``False`` otherwise. For schemas the property is checked on all global
         components. For XSD components check only the building of local subcomponents.
         """
-        raise NotImplementedError
+        raise NotImplementedError()
 
     @property
     def validation_attempted(self):
         """
-        Property that returns the *validation status* of the XSD validator. It can be 'full', 'partial' or 'none'.
+        Property that returns the *validation status* of the XSD validator.
+        It can be 'full', 'partial' or 'none'.
 
         | https://www.w3.org/TR/xmlschema-1/#e-validation_attempted
         | https://www.w3.org/TR/2012/REC-xmlschema11-1-20120405/#e-validation_attempted
         """
-        raise NotImplementedError
+        raise NotImplementedError()
 
     @property
     def validity(self):
         """
-        Property that returns the XSD validator's validity. It can be ‘valid’, ‘invalid’ or ‘notKnown’.
+        Property that returns the XSD validator's validity.
+        It can be ‘valid’, ‘invalid’ or ‘notKnown’.
 
         | https://www.w3.org/TR/xmlschema-1/#e-validity
         | https://www.w3.org/TR/2012/REC-xmlschema11-1-20120405/#e-validity
         """
-        if self.errors or any(comp.errors for comp in self.iter_components()):
+        if self.validation == 'skip':
+            return 'notKnown'
+        elif self.errors or any(comp.errors for comp in self.iter_components()):
             return 'invalid'
         elif self.built:
             return 'valid'
@@ -106,7 +105,7 @@ class XsdValidator(object):
         :param xsd_classes: returns only a specific class/classes of components, \
         otherwise returns all components.
         """
-        raise NotImplementedError
+        raise NotImplementedError()
 
     @property
     def all_errors(self):
@@ -137,8 +136,11 @@ class XsdValidator(object):
         attribute of the validator, if it's present.
         :param validation: overrides the default validation mode of the validator.
         """
-        if validation not in XSD_VALIDATION_MODES:
+        if validation:
+            check_validation_mode(validation)
+        else:
             validation = self.validation
+
         if validation == 'skip':
             return
 
@@ -147,7 +149,8 @@ class XsdValidator(object):
         elif elem is None:
             elem = getattr(self, 'elem', None)
         else:
-            raise XMLSchemaValueError("'elem' argument must be an Element instance, not %r." % elem)
+            msg = "the argument 'elem' must be an Element instance, not {!r}."
+            raise XMLSchemaTypeError(msg.format(elem))
 
         if isinstance(error, XMLSchemaParseError):
             error.validator = self
@@ -155,14 +158,15 @@ class XsdValidator(object):
             error.elem = elem
             error.source = getattr(self, 'source', None)
         elif isinstance(error, Exception):
-            message = unicode_type(error).strip()
+            message = str(error).strip()
             if message[0] in '\'"' and message[0] == message[-1]:
                 message = message.strip('\'"')
             error = XMLSchemaParseError(self, message, elem)
-        elif isinstance(error, string_base_type):
+        elif isinstance(error, str):
             error = XMLSchemaParseError(self, error, elem)
         else:
-            raise XMLSchemaValueError("'error' argument must be an exception or a string, not %r." % error)
+            msg = "'error' argument must be an exception or a string, not {!r}."
+            raise XMLSchemaTypeError(msg.format(error))
 
         if validation == 'lax':
             self.errors.append(error)
@@ -202,10 +206,12 @@ class XsdComponent(XsdValidator):
 
     :param elem: ElementTree's node containing the definition.
     :param schema: the XMLSchema object that owns the definition.
-    :param parent: the XSD parent, `None` means that is a global component that has the schema as parent.
+    :param parent: the XSD parent, `None` means that is a global component that \
+    has the schema as parent.
     :param name: name of the component, maybe overwritten by the parse of the `elem` argument.
 
-    :cvar qualified: for name matching, unqualified matching may be admitted only for elements and attributes.
+    :cvar qualified: for name matching, unqualified matching may be admitted only \
+    for elements and attributes.
     :vartype qualified: bool
     """
     _REGEX_SPACE = re.compile(r'\s')
@@ -231,7 +237,9 @@ class XsdComponent(XsdValidator):
     def __setattr__(self, name, value):
         if name == "elem":
             if not is_etree_element(value):
-                raise XMLSchemaTypeError("%r attribute must be an Etree Element: %r" % (name, value))
+                raise XMLSchemaTypeError(
+                    "%r attribute must be an Etree Element: %r" % (name, value)
+                )
             elif value.tag not in self._ADMITTED_TAGS:
                 raise XMLSchemaValueError(
                     "wrong XSD element %r for %r, must be one of %r." % (
@@ -341,12 +349,13 @@ class XsdComponent(XsdValidator):
             elif self.parent is None:
                 self.parse_error("missing attribute 'name' in a global %r" % type(self))
             else:
-                self.parse_error("missing both attributes 'name' and 'ref' in local %r" % type(self))
-            self.name = 'nameless_%s' % str(id(self))
-        elif self.parent is None:
-            self.parse_error("attribute 'ref' not allowed in a global %r" % type(self))
+                self.parse_error(
+                    "missing both attributes 'name' and 'ref' in local %r" % type(self)
+                )
         elif 'name' in self.elem.attrib:
             self.parse_error("attributes 'name' and 'ref' are mutually exclusive")
+        elif self.parent is None:
+            self.parse_error("attribute 'ref' not allowed in a global %r" % type(self))
         else:
             try:
                 self.name = self.schema.resolve_qname(ref)
@@ -354,7 +363,8 @@ class XsdComponent(XsdValidator):
                 self.parse_error(err)
             else:
                 if self._parse_child_component(self.elem) is not None:
-                    self.parse_error("a reference component cannot has child definitions/declarations")
+                    self.parse_error("a reference component cannot have "
+                                     "child definitions/declarations")
                 return True
 
     def _parse_boolean_attribute(self, name):
@@ -372,7 +382,7 @@ class XsdComponent(XsdValidator):
 
     def _parse_child_component(self, elem, strict=True):
         child = None
-        for index, child in enumerate(filter(lambda x: x.tag != XSD_ANNOTATION, elem)):
+        for index, child in enumerate(filter(is_not_xsd_annotation, elem)):
             if not strict:
                 return child
             elif index:
@@ -389,23 +399,25 @@ class XsdComponent(XsdValidator):
 
         self._target_namespace = self.elem.attrib['targetNamespace'].strip()
         if 'name' not in self.elem.attrib:
-            self.parse_error("attribute 'name' must be present when 'targetNamespace' attribute is provided")
+            self.parse_error("attribute 'name' must be present when "
+                             "'targetNamespace' attribute is provided")
         if 'form' in self.elem.attrib:
-            self.parse_error("attribute 'form' must be absent when 'targetNamespace' attribute is provided")
+            self.parse_error("attribute 'form' must be absent when "
+                             "'targetNamespace' attribute is provided")
         if self._target_namespace != self.schema.target_namespace:
             if self.parent is None:
-                self.parse_error("a global attribute must has the same namespace as its parent schema")
+                self.parse_error("a global %s must have the same namespace as "
+                                 "its parent schema" % self.__class__.__name__)
 
             xsd_type = self.get_parent_type()
-            if xsd_type and xsd_type.parent is None and \
-                    (xsd_type.derivation != 'restriction' or xsd_type.base_type.name == XSD_ANY_TYPE):
+            if not xsd_type or xsd_type.parent is not None:
+                pass
+            elif xsd_type.derivation != 'restriction' or xsd_type.base_type.name == XSD_ANY_TYPE:
                 self.parse_error("a declaration contained in a global complexType "
-                                 "must has the same namespace as its parent schema")
+                                 "must have the same namespace as its parent schema")
 
-        if not self._target_namespace and self.name[0] == '{':
+        if not self._target_namespace:
             self.name = local_name(self.name)
-        elif self.name[0] != '{':
-            self.name = '{%s}%s' % (self._target_namespace, self.name)
         else:
             self.name = '{%s}%s' % (self._target_namespace, local_name(self.name))
 
@@ -432,7 +444,7 @@ class XsdComponent(XsdValidator):
 
     @property
     def built(self):
-        raise NotImplementedError
+        raise NotImplementedError()
 
     def is_matching(self, name, default_namespace=None, **kwargs):
         """
@@ -455,7 +467,10 @@ class XsdComponent(XsdValidator):
             return self.qualified_name == qname or not self.qualified and self.local_name == name
 
     def match(self, name, default_namespace=None, **kwargs):
-        """Returns the component if its name is matching the name provided as argument, `None` otherwise."""
+        """
+        Returns the component if its name is matching the name provided as argument,
+        `None` otherwise.
+        """
         return self if self.is_matching(name, default_namespace, **kwargs) else None
 
     def get_global(self):
@@ -508,11 +523,7 @@ class XsdComponent(XsdValidator):
                 break
 
     def tostring(self, indent='', max_lines=None, spaces_for_tab=4):
-        """
-        Returns the XML elements that declare or define the component as a string.
-        """
-        if self.elem is None:
-            return str(None)  # Incomplete component
+        """Serializes the XML elements that declare or define the component to a string."""
         return etree_tostring(self.schema_elem, self.namespaces, indent, max_lines, spaces_for_tab)
 
 
@@ -581,7 +592,7 @@ class XsdType(XsdComponent):
 
     @property
     def built(self):
-        raise NotImplementedError
+        raise NotImplementedError()
 
     @property
     def content_type_label(self):
@@ -614,12 +625,12 @@ class XsdType(XsdComponent):
     @staticmethod
     def is_simple():
         """Returns `True` if the instance is a simpleType, `False` otherwise."""
-        raise NotImplementedError
+        raise NotImplementedError()
 
     @staticmethod
     def is_complex():
         """Returns `True` if the instance is a complexType, `False` otherwise."""
-        raise NotImplementedError
+        raise NotImplementedError()
 
     @staticmethod
     def is_atomic():
@@ -635,33 +646,40 @@ class XsdType(XsdComponent):
 
     def is_empty(self):
         """Returns `True` if the instance has an empty value or content, `False` otherwise."""
-        raise NotImplementedError
+        raise NotImplementedError()
 
     def is_emptiable(self):
         """Returns `True` if the instance has an emptiable value or content, `False` otherwise."""
-        raise NotImplementedError
+        raise NotImplementedError()
 
     def has_simple_content(self):
         """
         Returns `True` if the instance is a simpleType or a complexType with simple
         content, `False` otherwise.
         """
-        raise NotImplementedError
+        raise NotImplementedError()
 
     def has_mixed_content(self):
         """
         Returns `True` if the instance is a complexType with mixed content, `False` otherwise.
         """
-        raise NotImplementedError
+        raise NotImplementedError()
 
     def is_element_only(self):
         """
-        Returns `True` if the instance is a complexType with element-only content, `False` otherwise.
+        Returns `True` if the instance is a complexType with element-only content,
+        `False` otherwise.
         """
-        raise NotImplementedError
+        raise NotImplementedError()
 
     def is_derived(self, other, derivation=None):
-        raise NotImplementedError
+        raise NotImplementedError()
+
+    def is_extension(self):
+        return self.derivation == 'extension'
+
+    def is_restriction(self):
+        return self.derivation == 'restriction'
 
     def is_blocked(self, xsd_element):
         """
@@ -685,8 +703,11 @@ class XsdType(XsdComponent):
     def is_key(self):
         return self.name == XSD_ID or self.is_derived(self.maps.types[XSD_ID])
 
+    def is_qname(self):
+        return self.name == XSD_QNAME or self.is_derived(self.maps.types[XSD_QNAME])
+
     def text_decode(self, text):
-        raise NotImplementedError
+        raise NotImplementedError()
 
 
 class ValidationMixin(object):
@@ -721,8 +742,9 @@ class ValidationMixin(object):
         :param use_defaults: indicates whether to use default values for filling missing data.
         :param namespaces: is an optional mapping from namespace prefix to URI.
         """
-        error = next(self.iter_errors(source, use_defaults=use_defaults, namespaces=namespaces), None)
-        return error is None
+        return next(
+            self.iter_errors(source, use_defaults=use_defaults, namespaces=namespaces), None
+        ) is None
 
     def iter_errors(self, source, use_defaults=True, namespaces=None):
         """
@@ -759,18 +781,19 @@ class ValidationMixin(object):
         :raises: :exc:`XMLSchemaValidationError` if the object is not decodable by \
         the XSD component, or also if it's invalid when ``validation='strict'`` is provided.
         """
-        if validation not in XSD_VALIDATION_MODES:
-            raise XMLSchemaValueError("validation argument can be 'strict', 'lax' or 'skip': %r" % validation)
+        check_validation_mode(validation)
 
-        errors = []
+        result, errors = None, []
         for result in self.iter_decode(source, validation, **kwargs):
             if not isinstance(result, XMLSchemaValidationError):
-                return (result, errors) if validation == 'lax' else result
+                break
             elif validation == 'strict':
                 raise result
             elif validation == 'lax':
                 errors.append(result)
 
+        return (result, errors) if validation == 'lax' else result
+
     def encode(self, obj, validation='strict', **kwargs):
         """
         Encodes data to XML.
@@ -785,18 +808,19 @@ class ValidationMixin(object):
         :raises: :exc:`XMLSchemaValidationError` if the object is not encodable by the XSD \
         component, or also if it's invalid when ``validation='strict'`` is provided.
         """
-        if validation not in XSD_VALIDATION_MODES:
-            raise XMLSchemaValueError("validation argument can be 'strict', 'lax' or 'skip': %r" % validation)
+        check_validation_mode(validation)
 
-        errors = []
+        result, errors = None, []
         for result in self.iter_encode(obj, validation=validation, **kwargs):
             if not isinstance(result, XMLSchemaValidationError):
-                return (result, errors) if validation == 'lax' else result
+                break
             elif validation == 'strict':
                 raise result
             elif validation == 'lax':
                 errors.append(result)
 
+        return (result, errors) if validation == 'lax' else result
+
     def iter_decode(self, source, validation='lax', **kwargs):
         """
         Creates an iterator for decoding an XML source to a Python object.
@@ -807,7 +831,7 @@ class ValidationMixin(object):
         :return: Yields a decoded object, eventually preceded by a sequence of \
         validation or decoding errors.
         """
-        raise NotImplementedError
+        raise NotImplementedError()
 
     def iter_encode(self, obj, validation='lax', **kwargs):
         """
@@ -819,12 +843,13 @@ class ValidationMixin(object):
         :return: Yields an Element, eventually preceded by a sequence of validation \
         or encoding errors.
         """
-        raise NotImplementedError
+        raise NotImplementedError()
 
-    def validation_error(self, validation, error, obj=None, source=None, namespaces=None, **_kwargs):
+    def validation_error(self, validation, error, obj=None,
+                         source=None, namespaces=None, **_kwargs):
         """
-        Helper method for generating and updating validation errors. Incompatible with 'skip'
-        validation mode. Il validation mode is 'lax' returns the error, otherwise raises the error.
+        Helper method for generating and updating validation errors. If validation
+        mode is 'lax' or 'skip' returns the error, otherwise raises the error.
 
         :param validation: an error-compatible validation mode: can be 'lax' or 'strict'.
         :param error: an error instance or the detailed reason of failed validation.
@@ -833,9 +858,8 @@ class ValidationMixin(object):
         :param namespaces: is an optional mapping from namespace prefix to URI.
         :param _kwargs: keyword arguments of the validation process that are not used.
         """
-        if not isinstance(error, XMLSchemaValidationError):
-            error = XMLSchemaValidationError(self, obj, error, source, namespaces)
-        else:
+        check_validation_mode(validation)
+        if isinstance(error, XMLSchemaValidationError):
             if error.namespaces is None and namespaces is not None:
                 error.namespaces = namespaces
             if error.source is None and source is not None:
@@ -844,61 +868,14 @@ class ValidationMixin(object):
                 error.obj = obj
             if error.elem is None and is_etree_element(obj):
                 error.elem = obj
-
-        if validation == 'lax':
-            return error
-        elif validation == 'strict':
-            raise error
-        elif validation == 'skip':
-            raise XMLSchemaValueError("validation mode 'skip' incompatible with error generation.")
+        elif isinstance(error, Exception):
+            error = XMLSchemaValidationError(self, obj, str(error), source, namespaces)
         else:
-            raise XMLSchemaValueError("unknown validation mode %r" % validation)
+            error = XMLSchemaValidationError(self, obj, error, source, namespaces)
 
-    def decode_error(self, validation, obj, decoder, reason=None, source=None, namespaces=None, **_kwargs):
-        """
-        Helper method for generating decode errors. Incompatible with 'skip' validation mode.
-        Il validation mode is 'lax' returns the error, otherwise raises the error.
-
-        :param validation: an error-compatible validation mode: can be 'lax' or 'strict'.
-        :param obj: the not validated XML data.
-        :param decoder: the XML data decoder.
-        :param reason: the detailed reason of failed validation.
-        :param source: the XML resource that contains the error.
-        :param namespaces: is an optional mapping from namespace prefix to URI.
-        :param _kwargs: keyword arguments of the validation process that are not used.
-        """
-        error = XMLSchemaDecodeError(self, obj, decoder, reason, source, namespaces)
-        if validation == 'lax':
-            return error
-        elif validation == 'strict':
+        if validation == 'strict' and error.elem is not None:
             raise error
-        elif validation == 'skip':
-            raise XMLSchemaValueError("validation mode 'skip' incompatible with error generation.")
-        else:
-            raise XMLSchemaValueError("unknown validation mode %r" % validation)
-
-    def encode_error(self, validation, obj, encoder, reason=None, source=None, namespaces=None, **_kwargs):
-        """
-        Helper method for generating encode errors. Incompatible with 'skip' validation mode.
-        Il validation mode is 'lax' returns the error, otherwise raises the error.
-
-        :param validation: an error-compatible validation mode: can be 'lax' or 'strict'.
-        :param obj: the not validated XML data.
-        :param encoder: the XML encoder.
-        :param reason: the detailed reason of failed validation.
-        :param source: the XML resource that contains the error.
-        :param namespaces: is an optional mapping from namespace prefix to URI.
-        :param _kwargs: keyword arguments of the validation process that are not used.
-        """
-        error = XMLSchemaEncodeError(self, obj, encoder, reason, source, namespaces)
-        if validation == 'lax':
-            return error
-        elif validation == 'strict':
-            raise error
-        elif validation == 'skip':
-            raise XMLSchemaValueError("validation mode 'skip' incompatible with error generation.")
-        else:
-            raise XMLSchemaValueError("unknown validation mode %r" % validation)
+        return error
 
 
 class ParticleMixin(object):
@@ -915,6 +892,14 @@ class ParticleMixin(object):
     def occurs(self):
         return [self.min_occurs, self.max_occurs]
 
+    @property
+    def effective_min_occurs(self):
+        return self.min_occurs
+
+    @property
+    def effective_max_occurs(self):
+        return self.max_occurs
+
     def is_emptiable(self):
         return self.min_occurs == 0
 
@@ -948,19 +933,8 @@ class ParticleMixin(object):
         else:
             return self.max_occurs <= other.max_occurs
 
-    @property
-    def effective_min_occurs(self):
-        return self.min_occurs
-
-    @property
-    def effective_max_occurs(self):
-        return self.max_occurs
-
-    ###
-    # Methods used by XSD components
-    def parse_error(self, *args, **kwargs):
-        """Helper method overridden by XsdValidator.parse_error() in XSD components."""
-        raise XMLSchemaParseError(*args)
+    def parse_error(self, message):
+        raise XMLSchemaParseError(self, message)
 
     def _parse_particle(self, elem):
         if 'minOccurs' in elem.attrib:
