@@ -1,6 +1,5 @@
-# -*- coding: utf-8 -*-
 #
-# Copyright (c), 2016-2019, SISSA (International School for Advanced Studies).
+# Copyright (c), 2016-2020, SISSA (International School for Advanced Studies).
 # All rights reserved.
 # This file is distributed under the terms of the MIT License.
 # See the file 'LICENSE' in the root directory of the present
@@ -11,29 +10,29 @@
 """
 This module defines a mixin class for enabling XPath on schemas.
 """
-from __future__ import unicode_literals
 from abc import abstractmethod
+from collections.abc import Sequence
 from elementpath import XPath2Parser, XPathSchemaContext, AbstractSchemaProxy
+import re
 import threading
 
-from .compat import Sequence
 from .qnames import XSD_SCHEMA
 from .namespaces import XSD_NAMESPACE
 from .exceptions import XMLSchemaValueError, XMLSchemaTypeError
 
+_REGEX_TAG_POSITION = re.compile(r'\b\[\d+\]')
+
 
 class XMLSchemaContext(XPathSchemaContext):
     """
-    XPath dynamic context class for *xmlschema* library. Implements safe iteration
+    XPath dynamic context class for *xmlschema* library. Skips elem.text,
+    that is always `None` for schema elements, and implements safe iteration
     methods for schema elements that recognize circular references.
     """
     def _iter_descendants(self):
         def safe_iter_descendants(context):
             elem = context.item
             yield elem
-            if elem.text is not None:
-                context.item = elem.text
-                yield context.item
             if len(elem):
                 context.size = len(elem)
                 for context.position, context.item in enumerate(elem):
@@ -54,9 +53,6 @@ class XMLSchemaContext(XPathSchemaContext):
         def safe_iter_context(context):
             elem = context.item
             yield elem
-            if elem.text is not None:
-                context.item = elem.text
-                yield context.item
 
             for item in elem.attrib.items():
                 context.item = item
@@ -66,14 +62,12 @@ class XMLSchemaContext(XPathSchemaContext):
                 context.size = len(elem)
                 for context.position, context.item in enumerate(elem):
                     if context.item.parent is None:
-                        for item in safe_iter_context(context):
-                            yield item
+                        yield from safe_iter_context(context)
                     elif getattr(context.item, 'ref', None) is not None:
                         yield context.item
                     elif context.item not in local_items:
                         local_items.append(context.item)
-                        for item in safe_iter_context(context):
-                            yield item
+                        yield from safe_iter_context(context)
 
         local_items = []
         return safe_iter_context(self)
@@ -172,8 +166,8 @@ class ElementPathMixin(Sequence):
     """
     Mixin abstract class for enabling ElementTree and XPath API on XSD components.
 
-    :cvar text: The Element text. Its value is always `None`. For compatibility with the ElementTree API.
-    :cvar tail: The Element tail. Its value is always `None`. For compatibility with the ElementTree API.
+    :cvar text: the Element text, for compatibility with the ElementTree API.
+    :cvar tail: the Element tail, for compatibility with the ElementTree API.
     """
     text = None
     tail = None
@@ -199,12 +193,12 @@ class ElementPathMixin(Sequence):
 
     @abstractmethod
     def __iter__(self):
-        pass
+        raise NotImplementedError
 
     def __getitem__(self, i):
         try:
             return [e for e in self][i]
-        except AttributeError:
+        except IndexError:
             raise IndexError('child index out of range')
 
     def __reversed__(self):
@@ -253,6 +247,7 @@ class ElementPathMixin(Sequence):
         path = path.strip()
         if path.startswith('/') and not path.startswith('//'):
             path = ''.join(['/', XSD_SCHEMA, path])
+        path = _REGEX_TAG_POSITION.sub('', path)  # Strips tags's positions from path
 
         namespaces = self._get_xpath_namespaces(namespaces)
         with self._xpath_lock:
@@ -310,15 +305,13 @@ class ElementPathMixin(Sequence):
                 yield elem
             for child in elem:
                 if child.parent is None:
-                    for e in safe_iter(child):
-                        yield e
+                    yield from safe_iter(child)
                 elif getattr(child, 'ref', None) is not None:
-                    if tag is None or elem.is_matching(tag):
+                    if tag is None or child.is_matching(tag):
                         yield child
                 elif child not in local_elements:
                     local_elements.append(child)
-                    for e in safe_iter(child):
-                        yield e
+                    yield from safe_iter(child)
 
         if tag == '*':
             tag = None
